#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from queue import Queue
from random import choice
from re import fullmatch
from string import ascii_lowercase
from threading import Thread
from time import sleep, time

from thingsboard_gateway.gateway.entities.converted_data import ConvertedData
from thingsboard_gateway.gateway.statistics.statistics_service import StatisticsService
from thingsboard_gateway.tb_utility.tb_loader import TBModuleLoader
from thingsboard_gateway.tb_utility.tb_utility import TBUtility
from thingsboard_gateway.tb_utility.tb_logger import init_logger

try:
    from requests import Timeout, request
except ImportError:
    print("Requests library not found - installing...")
    TBUtility.install_package("requests")
    from requests import Timeout, request
from requests.auth import HTTPBasicAuth
from requests.exceptions import RequestException, JSONDecodeError

from thingsboard_gateway.connectors.connector import Connector
from thingsboard_gateway.connectors.request.json_request_uplink_converter import JsonRequestUplinkConverter
from thingsboard_gateway.connectors.request.json_request_downlink_converter import JsonRequestDownlinkConverter


class RequestConnector(Connector, Thread):
    def __init__(self, gateway, config, connector_type):
        super().__init__()
        self.__rpc_requests = []
        self.__config = config
        self.__id = self.__config.get('id')
        self._connector_type = connector_type
        self.__gateway = gateway
        self.name = self.__config.get("name", "".join(choice(ascii_lowercase) for _ in range(5)))
        self._log = init_logger(self.__gateway, self.name, self.__config.get('logLevel', 'INFO'),
                                enable_remote_logging=self.__config.get('enableRemoteLogging', False),
                                is_connector_logger=True)
        self._converter_log = init_logger(self.__gateway, self.name + '_converter',
                                          self.__config.get('logLevel', 'INFO'),
                                          enable_remote_logging=self.__config.get('enableRemoteLogging', False),
                                          is_converter_logger=True, attr_name=self.name)
        self.__security = HTTPBasicAuth(self.__config["security"]["username"], self.__config["security"]["password"]) if \
            self.__config["security"]["type"] == "basic" else None
        self.__host = None
        self.__service_headers = {}
        if "http://" in self.__config["host"].lower() or "https://" in self.__config["host"].lower():
            self.__host = self.__config["host"]
        else:
            self.__host = "http://" + self.__config["host"]
        self.__ssl_verify = self.__config.get("SSLVerify", False)
        self.daemon = True
        self.__connected = False
        self.__stopped = False
        self.__requests_in_progress = []
        self.__convert_queue = Queue(1000000)
        self.__attribute_updates = []
        self.__fill_attribute_updates()
        self.__fill_rpc_requests()
        self.__fill_requests()

    def run(self):
        while not self.__stopped:
            request_sent = False
            if self.__requests_in_progress:
                for req in self.__requests_in_progress:
                    if time() >= req["next_time"]:
                        thread = Thread(target=self.__send_request, args=(req, self.__convert_queue, self._log),
                                        daemon=True,
                                        name="Request to endpoint \'%s\' Thread" % (req["config"].get("url")))
                        thread.start()
                        request_sent = True
            if not request_sent:
                sleep(.2)
            self.__process_data()

    def on_attributes_update(self, content):
        try:
            for attribute_request in self.__attribute_updates:
                if fullmatch(attribute_request["deviceNameFilter"], content["device"]) and fullmatch(
                        attribute_request["attributeFilter"], list(content["data"].keys())[0]):
                    converted_data = attribute_request["converter"].convert(attribute_request, content)
                    response_queue = Queue(1)
                    request_dict = {"config": {**attribute_request,
                                               **converted_data},
                                    "request": request,
                                    "withResponse": True}
                    attribute_update_request_thread = Thread(target=self.__send_request,
                                                             args=(request_dict, response_queue, self._log),
                                                             daemon=True,
                                                             name="Attribute request to %s" % (converted_data["url"]))
                    attribute_update_request_thread.start()
                    attribute_update_request_thread.join()
                    if not response_queue.empty():
                        response = response_queue.get_nowait()
                        self._log.debug(response)
                    del response_queue
        except Exception as e:
            self._log.exception(e)

    def server_side_rpc_handler(self, content):
        try:
            # check if RPC method is reserved get/set
            self.__check_and_process_reserved_rpc(content)

            for rpc_request in self.__rpc_requests:
                if fullmatch(rpc_request["deviceNameFilter"], content["device"]) and fullmatch(
                        rpc_request["methodFilter"], content["data"]["method"]):
                    self.__process_rpc(rpc_request, content)
        except Exception as e:
            self._log.exception(e)

    def __process_rpc(self, rpc_request, content):
        converted_data = rpc_request["converter"].convert(rpc_request, content)
        response_queue = Queue(1)
        request_dict = {"config": {**rpc_request,
                                   **converted_data},
                        "request": request,
                        "withResponse": True}
        rpc_request_thread = Thread(target=self.__send_request,
                                    args=(request_dict, response_queue, self._log),
                                    daemon=True,
                                    name="RPC request to %s" % (converted_data["url"]))
        rpc_request_thread.start()
        rpc_request_thread.join()
        if not response_queue.empty():
            response = response_queue.get_nowait()

            if rpc_request.get('responseValueExpression'):
                response_value_expression = rpc_request['responseValueExpression']
                values = TBUtility.get_values(response_value_expression, response.json(),
                                              expression_instead_none=True)
                values_tags = TBUtility.get_values(
                    response_value_expression, response.json(), get_tag=True)
                full_value = response_value_expression
                for (value, value_tag) in zip(values, values_tags):
                    is_valid_value = "${" in response_value_expression and "}" in response_value_expression

                    full_value = full_value.replace('${' + str(value_tag) + '}', str(value)) if is_valid_value else str(value)

                self.__gateway.send_rpc_reply(device=content["device"], req_id=content["data"]["id"],
                                              content={'result': full_value})
                del response_queue
                return

            self.__gateway.send_rpc_reply(device=content["device"], req_id=content["data"]["id"],
                                          content={'result': response.text})
            del response_queue
            return

        self.__gateway.send_rpc_reply(device=content["device"], req_id=content["data"]["id"],
                                      success_sent=True)

        del response_queue

    def __check_and_process_reserved_rpc(self, content):
        rpc_method_name = content["data"]["method"]

        if rpc_method_name == 'get' or rpc_method_name == 'set':
            params = self.__parse_reserved_rpc_params(rpc_method_name, content["data"]["params"])

            rpc_request = self.__format_rpc_reqeust(params)

            rpc_request['converter'] = JsonRequestDownlinkConverter(rpc_request, self._converter_log)

            self.__process_rpc(rpc_request, content)

    def __parse_reserved_rpc_params(self, rpc_method_name, params):
        result_params = {}
        for param in params.split(';'):
            try:
                (key, value) = param.split('=')
            except ValueError:
                continue

            if key and value:
                result_params[key] = value

        if rpc_method_name == 'set':
            result_params['requestValueExpression'] = result_params.pop('value', None)

        return result_params

    def __format_rpc_reqeust(self, params):
        return {
            'requestUrlExpression': params['requestUrlExpression'],
            'responseTimeout': params.get('responseTimeout', 1),
            'httpMethod': params.get('httpMethod', 'GET'),
            'requestValueExpression': params.get('requestValueExpression', '${params}'),
            'responseValueExpression': params.get('responseValueExpression', None),
            'timeout': params.get('timeout', 0.5),
            'tries': params.get('tries', 3),
            'httpHeaders': params.get('httpHeaders', {
                'Content-Type': 'application/json'
            }),
        }

    def __fill_requests(self):
        self._log.debug(self.__config["mapping"])
        for endpoint in self.__config["mapping"]:
            try:
                self._log.debug(endpoint)
                converter = None
                if endpoint["converter"]["type"] == "custom":
                    module = TBModuleLoader.import_module(self._connector_type, endpoint["converter"]["extension"])
                    if module:
                        self._log.debug('Custom converter for url %s - found!', endpoint["url"])
                        converter = module(endpoint, self._converter_log)
                    else:
                        self._log.error(
                            "\n\nCannot find extension module for %s url.\nPlease check your configuration.\n",
                            endpoint["url"])
                else:
                    converter = JsonRequestUplinkConverter(endpoint, self._log)
                self.__requests_in_progress.append({"config": endpoint,
                                                    "converter": converter,
                                                    "next_time": time(),
                                                    "request": request})
            except Exception as e:
                self._log.exception(e)

    def __fill_attribute_updates(self):
        for attribute_request in self.__config.get("attributeUpdates", []):
            if attribute_request.get("converter") is not None:
                converter = TBModuleLoader.import_module("request", attribute_request["converter"])(attribute_request,
                                                                                                    self._converter_log)
            else:
                converter = JsonRequestDownlinkConverter(attribute_request, self._converter_log)
            attribute_request_dict = {**attribute_request, "converter": converter}
            self.__attribute_updates.append(attribute_request_dict)

    def __fill_rpc_requests(self):
        for rpc_request in self.__config.get("serverSideRpc", []):
            if rpc_request.get("converter") is not None:
                converter = TBModuleLoader.import_module("request", rpc_request["converter"])(rpc_request, self._converter_log)
            else:
                converter = JsonRequestDownlinkConverter(rpc_request, self._converter_log)
            rpc_request_dict = {**rpc_request, "converter": converter}
            self.__rpc_requests.append(rpc_request_dict)

    def __send_request(self, request, converter_queue, logger):
        url = ""
        try:
            request["next_time"] = time() + request["config"].get("scanPeriod", 10)
            if request.get("converter") is None and isinstance(request["config"].get("converter"), dict):
                logger.error("Converter for request to '%s' endpoint is not defined. Request will be skipped.", request["config"].get("url"))
                return
            request_url_from_config = request["config"]["url"]
            request_url_from_config = (
                str("/" + request_url_from_config)
                if not request_url_from_config.startswith("/")
                   and not request_url_from_config.startswith("http")
                else request_url_from_config
            )
            logger.debug("Obtained request url from config - %s ", request_url_from_config)
            url, response = self.__execute_request(request, request_url_from_config, logger)

            if request.get('withResponse'):
                converter_queue.put(response)
                return

            if response and response.ok:
                if not converter_queue.full():
                    config_converter_data = [url, request["converter"]]
                    try:
                        json_response = response.json()

                        # Unpack data if dataUnpackExpression is defined in config
                        # This allows to unpack JSON responses that have final data at a sub key on any level
                        # {
                        #   "device": [...]
                        # }
                        data_unpack_expression = request["config"].get("dataUnpackExpression")
                        if data_unpack_expression:
                            json_response = TBUtility.get_value(data_unpack_expression, json_response, value_type="json")

                        config_converter_data.append(json_response)
                    except UnicodeDecodeError:
                        config_converter_data.append(response.content)
                    except JSONDecodeError:
                        config_converter_data.append(response.content)

                    if len(config_converter_data) == 3:
                        # Process sub requests if defined in config
                        if request["config"].get("subRequests"):
                            self.__process_sub_requests(request, url, config_converter_data[2], logger)
                        self.__convert_data(config_converter_data)
            else:
                logger.error("Request to URL: %s finished with code: %i", url, response.status_code)
        except Timeout:
            logger.error("Timeout error on request %s.", url)
        except RequestException as e:
            logger.error("Cannot connect to %s. Connection error.", url)
            logger.debug(e)
        except ConnectionError:
            logger.error("Cannot connect to %s. Connection error.", url)
        except Exception as e:
            logger.exception(e)

    def __execute_request(self, request, request_url, logger):
        url = self.__host + request_url if not request_url.lower().startswith("http") else request_url

        request_timeout = request["config"].get("timeout", 1)
        params = {
            "method": request["config"].get("httpMethod", "GET"),
            "url": url,
            "timeout": request_timeout,
            "allow_redirects": request["config"].get("allowRedirects", False),
            "verify": self.__ssl_verify,
            "auth": self.__security,
            "data": request["config"].get("data", {})
        }
        logger.debug("Full url request has been formed - %s", url)

        if request["config"].get("httpHeaders") is not None:
            params["headers"] = request["config"]["httpHeaders"]

        logger.debug("Request to %s will be sent", url)
        if isinstance(params["data"], str):
            params["data"] = params["data"].encode("utf-8")
        response = request["request"](**params)

        return url, response

    def __convert_data(self, data):
        try:
            url, converter, data = data
            data_to_send = []

            StatisticsService.count_connector_message(self.name, stat_parameter_name='connectorMsgsReceived')
            StatisticsService.count_connector_bytes(self.name, data, stat_parameter_name='connectorBytesReceived')

            if isinstance(data, list):
                for data_item in data:
                    self.__add_ts(data_item)
                    converted_data = converter.convert(url, data_item)
                    data_to_send.append(converted_data)
            else:
                self.__add_ts(data)
                data_to_send.append(converter.convert(url, data))

            for to_send in data_to_send:
                self.__convert_queue.put(to_send)

        except Exception as e:
            self._log.exception(e)

    def __add_ts(self, data):
        if isinstance(data, list):
            for item in data:
                self.__add_ts(item)
        elif isinstance(data, dict):
            if data.get("ts") is None:
                data["ts"] = int(time() * 1000)

    def __process_data(self):
        try:
            if not self.__convert_queue.empty():
                data: ConvertedData = self.__convert_queue.get()
                if data and (data.attributes_datapoints_count > 0 or data.telemetry_datapoints_count > 0):
                    self.__gateway.send_to_storage(self.get_name(), self.get_id(), data)

        except Exception as e:
            self._log.exception(e)

    def get_id(self):
        return self.__id

    def get_name(self):
        return self.name

    def get_type(self):
        return self._connector_type

    def is_connected(self):
        return self.__connected

    def is_stopped(self):
        return self.__stopped

    def open(self):
        self.__stopped = False
        self.start()

    def close(self):
        self.__stopped = True
        self._log.info("%r has been stopped.", self.name)
        self._log.stop()

    def get_config(self):
        return self.__config

    def __process_sub_requests(self, request, url, data, logger):
        datatypes = {"attributes": "attributes",
                     "telemetry": "telemetry"}
        data = data if isinstance(data, list) else [data]

        for data_item in data:
            for datatype in datatypes:
                for datatype_object_config in request["config"]["converter"].get(datatype, []):
                    # Check if a sub request for key is needed
                    key = datatype_object_config.get("key")
                    if key in request["config"].get("subRequests", {}):
                        request_url_from_config = TBUtility.replace_params_tags(request["config"]["subRequests"][key]["url"],
                                                                                {"data": data_item})

                        if not request_url_from_config.lower().startswith("http"):
                            if not request_url_from_config.startswith("/"):
                                request_url_from_config = "/" + request_url_from_config
                            request_url_from_config = url + request_url_from_config
                        logger.debug("Sub request needed for key %s with url %s", key, request_url_from_config)

                        response = self.__send_sub_request(request, request_url_from_config, logger)
                        logger.debug("Sub request response: %s", response)

                        # Only if a response is available, process it
                        if response:
                            result = response
                            # Make processing function available if defined and call it
                            processing_function = request["config"]["subRequests"][key].get("processingFunction")
                            if processing_function:
                                logger.trace("Processing sub request response with function:\n%s", processing_function)
                                local_scope = {}
                                exec(processing_function, {}, local_scope)
                                result = local_scope["process_data"](response, key)
                            # Update data with result of sub request
                            data_item.update(result)
                            logger.debug("Data after sub request processing: %s", data_item)

    def __send_sub_request(self, request, sub_request_url, logger):
        url = ""
        try:
            url, response = self.__execute_request(request, sub_request_url, logger)
            if response and response.ok:
                try:
                    return response.json()
                except UnicodeDecodeError:
                    return response.content
                except JSONDecodeError:
                    return response.content
            else:
                logger.error("Request to URL: %s finished with code: %i", url, response.status_code)
        except Timeout:
            logger.error("Timeout error on request %s.", url)
        except RequestException as e:
            logger.error("Cannot connect to %s. Connection error.", url)
            logger.debug(e)
        except ConnectionError:
            logger.error("Cannot connect to %s. Connection error.", url)
        except Exception as e:
            logger.exception(e)



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import re

from asyncua.common.subscription import Subscription

from thingsboard_gateway.gateway.constants import REPORT_STRATEGY_PARAMETER
from thingsboard_gateway.gateway.entities.report_strategy_config import ReportStrategyConfig
from thingsboard_gateway.connectors.opcua.entities.rpc_request import OpcUaRpcRequest


class Device:
    ABSOLUTE_PATH_PATTERN = re.compile(r"\${(Root\\.[A-Za-z0-9_.:\\\[\]\-()]+)}")
    RELATIVE_PATH_PATTERN = re.compile(r"\${([A-Za-z0-9_.:\\\[\]\-()]+)}")
    NODE_ID_PATTERN = re.compile(r"(ns=\d+;[isgb]=[^}]+)")

    def __init__(self, path, name, device_profile, config, converter, converter_for_sub, device_node, logger):
        self._log = logger
        self.__configured_values_count = 0
        self.path = path
        self.device_node = device_node
        self.name = name
        self.device_profile = device_profile
        self.config = config
        self.converter = converter
        self.converter_for_sub = converter_for_sub
        self.values = {
            'timeseries': [],
            'attributes': []
        }
        self.shared_attributes_keys = self.__get_shared_attributes_keys()
        self.shared_attributes_keys_value_pairs = self.__match_key_value_for_attribute_updates()
        self.nodes = []
        self.subscription: Subscription | None = None
        self.nodes_data_change_subscriptions = {}
        self.report_strategy = None
        if self.config.get(REPORT_STRATEGY_PARAMETER):
            try:
                self.report_strategy = ReportStrategyConfig(self.config.get(REPORT_STRATEGY_PARAMETER))
            except ValueError as e:
                self._log.error('Invalid report strategy config for %s: %s, connector report strategy will be used',
                                self.name, e)

        self.load_values()

    def __get_shared_attributes_keys(self):
        result = []

        for attr_config in self.config.get('attributes_updates', []):
            result.append(attr_config['key'])

        return result

    def __match_key_value_for_attribute_updates(self):
        result = {}
        for attr_config in self.config.get('attributes_updates', []):
            result[attr_config['key']] = attr_config['value'] if attr_config['value'] else None
        return result

    def __repr__(self):
        return f'<Device> Path: {self.path}, Name: {self.name}, Configured values: {self.__configured_values_count}'

    def load_values(self):
        self.__configured_values_count = 0

        for section in ('attributes', 'timeseries'):
            for node_config in self.config.get(section, []):
                try:
                    value_str = node_config['value']

                    # Match NodeId value (e.g. ns=2;s=SomeNode)
                    node_id_match = re.search(Device.NODE_ID_PATTERN, value_str)
                    if node_id_match:
                        self.values[section].append({
                            'path': node_id_match.group(1),
                            'key': node_config['key'],
                            'timestampLocation': node_config.get('timestampLocation', 'gateway'),
                            REPORT_STRATEGY_PARAMETER: node_config.get(REPORT_STRATEGY_PARAMETER)
                        })
                        continue

                    # Match absolute path (e.g. ${Root\.Objects.Device.SomeNode})
                    absolute_path_match = re.search(Device.ABSOLUTE_PATH_PATTERN, value_str)
                    if absolute_path_match:
                        full_path = absolute_path_match.group(1).split('\\.')
                        self.values[section].append({
                            'path': full_path,
                            'key': node_config['key'],
                            'timestampLocation': node_config.get('timestampLocation', 'gateway'),
                            REPORT_STRATEGY_PARAMETER: node_config.get(REPORT_STRATEGY_PARAMETER)
                        })
                        continue

                    # Match relative path (e.g. ${Device.SomeNode})
                    relative_path_match = re.search(Device.RELATIVE_PATH_PATTERN, value_str)
                    if relative_path_match:
                        full_path = self.path + relative_path_match.group(1).split('\\.')
                        self.values[section].append({
                            'path': full_path,
                            'key': node_config['key'],
                            'timestampLocation': node_config.get('timestampLocation', 'gateway'),
                            REPORT_STRATEGY_PARAMETER: node_config.get(REPORT_STRATEGY_PARAMETER)
                        })
                        continue

                except KeyError as e:
                    self._log.error('Invalid config for %s (key %s not found)', node_config, e)

            self.__configured_values_count += len(self.values[section])

        self._log.debug('Loaded %r values for %s', len(self.values), self.name)

    def get_node_by_key(self, key):
        try:
            for node_config in self.nodes:
                if node_config['key'] == key:
                    return node_config['node']
            return None
        except KeyError:
            return None

    @staticmethod
    def is_valid_rpc_method_name(rpc_device_section: dict, rpc_request: OpcUaRpcRequest) -> bool:
        for rpc in rpc_device_section:
            if rpc.get('method') == rpc_request.rpc_method:
                return True
        return False

    @staticmethod
    def get_device_rpc_arguments(rpc_device_section: dict, rpc_request: OpcUaRpcRequest) -> list | None | dict:
        arguments = []
        found_method_rpc_section = next(
            filter(lambda rpc: rpc.get('method') == rpc_request.rpc_method, rpc_device_section))
        arguments_from_config = found_method_rpc_section.get('arguments', [])
        empty_arguments_condition = not arguments_from_config

        if empty_arguments_condition:
            return rpc_request.arguments

        if arguments_from_config:

            arguments = [argument['value'] for argument in arguments_from_config if argument.get('value')]

            if rpc_request.arguments and not isinstance(rpc_request.arguments, list):
                error_message = "The arguments must be specified in the square quotes []"
                return {"error": error_message}

            if rpc_request.arguments and len(rpc_request.arguments) == len(arguments_from_config):
                arguments = rpc_request.arguments
                return arguments

            elif arguments and len(arguments) != len(arguments_from_config):
                error_message = "You must either define values for arguments in config or along with rpc request"
                return {"error": error_message}

            elif (rpc_request.arguments and len(rpc_request.arguments) != len(arguments_from_config)) or (
                    not arguments and rpc_request.arguments is None):
                error_message = f"Expected {len(arguments_from_config)} arguments, but got {len(rpc_request.arguments) if rpc_request.arguments is not None else 0}"
                return {"error": error_message}

        return arguments




#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from thingsboard_gateway.connectors.knx.knx_converter import KNXConverter
from thingsboard_gateway.gateway.entities.converted_data import ConvertedData
from thingsboard_gateway.gateway.entities.report_strategy_config import ReportStrategyConfig
from thingsboard_gateway.gateway.statistics.statistics_service import StatisticsService
from thingsboard_gateway.tb_utility.tb_utility import TBUtility


class KNXUplinkConverter(KNXConverter):
    def __init__(self, config, logger):
        self.__log = logger
        self.__config = config

    def convert(self, data):
        StatisticsService.count_connector_message(self.__log.name, 'convertersMsgProcessed')
        device_name = self.__get_device_name(data)
        device_type = self.__get_device_type(data)

        converted_data = ConvertedData(device_name=device_name, device_type=device_type)

        device_report_strategy = self._get_device_report_strategy(self.__config.get('reportStrategy'),
                                                                  device_name)

        for section in ('attributes', 'timeseries'):
            for config in self.__config.get(section, []):
                try:
                    converted_value = data.get(config.get('groupAddress'))['response']
                    if converted_value is not None:
                        datapoint_key = TBUtility.convert_key_to_datapoint_key(config['key'],
                                                                               device_report_strategy,
                                                                               config,
                                                                               self.__log)

                        if section == 'attributes':
                            converted_data.add_to_attributes({datapoint_key: converted_value})
                        elif section == 'timeseries':
                            converted_data.add_to_telemetry({datapoint_key: converted_value})
                    else:
                        self.__log.warning('Received None value for group address %s', config.get('groupAddress'))
                except Exception as e:
                    self.__log.error('Error converting data: %s', e)

        StatisticsService.count_connector_message(self.__log.name,
                                                  'convertersAttrProduced',
                                                  count=converted_data.attributes_datapoints_count)
        StatisticsService.count_connector_message(self.__log.name,
                                                  'convertersTsProduced',
                                                  count=converted_data.telemetry_datapoints_count)

        self.__log.trace('Converted data: %s', data)
        return converted_data

    def __get_device_name(self, data):
        device_info = self.__config['deviceInfo']

        if device_info.get('deviceNameExpressionSource') == 'expression':
            group_address, converted_data = self.__find_converted_data_by_key('deviceName', data)
            return self.__replace_expression_with_value(device_info['deviceNameExpression'],
                                                        group_address,
                                                        converted_data)

        return device_info['deviceNameExpression']

    def __get_device_type(self, data):
        device_info = self.__config['deviceInfo']

        if device_info.get('deviceProfileExpressionSource') == 'expression':
            group_address, converted_data = self.__find_converted_data_by_key('deviceType', data)
            return self.__replace_expression_with_value(device_info['deviceTypeExpression'],
                                                        group_address,
                                                        converted_data)

        return device_info['deviceProfileNameExpression']

    @staticmethod
    def __find_converted_data_by_key(key, data):
        for group_address, value in data.items():
            if key in value['keys']:
                return group_address, value['response']

    @staticmethod
    def __replace_expression_with_value(expression, group_address, data):
        expression = expression.replace('${' + str(group_address) + '}', str(data))
        return expression

    def _get_device_report_strategy(self, report_strategy, device_name):
        try:
            return ReportStrategyConfig(report_strategy)
        except ValueError as e:
            self.__log.trace("Report strategy config is not specified for device %s: %s", device_name, e)



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import logging
from time import monotonic
from threading import RLock
from typing import TYPE_CHECKING
from os import path

if TYPE_CHECKING:
    from thingsboard_gateway.gateway.tb_gateway_service import TBGatewayService

from thingsboard_gateway.tb_utility.tb_rotating_file_handler import TimedRotatingFileHandler
from thingsboard_gateway.gateway.statistics.statistics_service import StatisticsService

TRACE_LOGGING_LEVEL = 5
logging.addLevelName(TRACE_LOGGING_LEVEL, "TRACE")


def init_logger(gateway: 'TBGatewayService', name, level, enable_remote_logging=False, is_connector_logger=False,
                is_converter_logger=False, attr_name=None):
    """
    For creating a Logger with all config automatically
    Create a Logger manually only if you know what you are doing!
    """

    log = logging.getLogger(name)
    log.is_connector_logger = is_connector_logger
    log.is_converter_logger = is_converter_logger

    if attr_name:
        log.attr_name = attr_name + '_ERRORS_COUNT'

    # Add file handler to the connector or converter logger
    # First check if it is a main module logger (for example OPC-UA connector logger)
    # If it is, add a file handler to the main module logger
    # If it is not (for example asyncua logger), add the main module file handler to the logger
    if TbLogger.is_main_module_logger(name, attr_name, is_converter_logger):
        file_handler = None

        if is_connector_logger:
            file_handler = TimedRotatingFileHandler.get_connector_file_handler(log.name + '_connector')

        if is_converter_logger:
            file_handler = TimedRotatingFileHandler.get_converter_file_handler(log.name)

        if file_handler:
            log.addHandler(file_handler)
    else:
        main_file_handler = TimedRotatingFileHandler.get_time_rotating_file_handler_by_logger_name(attr_name)
        if main_file_handler and main_file_handler not in log.handlers:
            log.addHandler(main_file_handler)

    from thingsboard_gateway.tb_utility.tb_handler import TBRemoteLoggerHandler
    if not hasattr(gateway, 'remote_handler'):
        gateway.remote_handler = TBRemoteLoggerHandler(gateway)
    if gateway.remote_handler not in log.handlers:
        log.addHandler(gateway.remote_handler)
        gateway.remote_handler.add_logger(name, level if enable_remote_logging else 'NONE')

    log_level_conf = level
    if log_level_conf:
        log_level = logging.getLevelName(log_level_conf)

        try:
            log.setLevel(log_level)
        except ValueError:
            log.setLevel(100)

    if log.is_connector_logger or log.is_converter_logger:
        level = level if enable_remote_logging else 'NONE'
        if gateway.remote_handler.loggers.get(name):
            gateway.remote_handler.update_logger(name, level)
        else:
            gateway.remote_handler.add_logger(name, log_level)

    return log


class TbLogger(logging.Logger):
    ALL_ERRORS_COUNT = 0
    IS_ALL_ERRORS_COUNT_RESET = False
    RESET_ERRORS_PERIOD = 60
    SEND_ERRORS_PERIOD = 5

    __PREVIOUS_ALL_ERRORS_COUNT = -1
    __PREVIOUS_BATCH_TO_SEND = {}

    ERRORS_MUTEX = RLock()
    ERRORS_BATCH = {}

    PREVIOUS_ERRORS_SENT_TIME = 0
    PREVIOUS_ERRORS_RESET_TIME = 0

    def __init__(self, name, level=logging.NOTSET, is_connector_logger=False, is_converter_logger=False,
                 attr_name=None):
        super(TbLogger, self).__init__(name=name, level=level)
        self.propagate = True
        self.parent = self.root
        self.__is_connector_logger = is_connector_logger
        self.__is_converter_logger = is_converter_logger
        self.__previous_reset_errors_time = TbLogger.PREVIOUS_ERRORS_RESET_TIME
        logging.Logger.trace = TbLogger.trace

        self.errors = 0

        if attr_name:
            self.attr_name = attr_name + '_ERRORS_COUNT'
        else:
            self.attr_name = self.name + '_ERRORS_COUNT'

        self._is_on_init_state = True

    @property
    def is_connector_logger(self):
        return self.__is_connector_logger

    @is_connector_logger.setter
    def is_connector_logger(self, value):
        self.__is_connector_logger = value

    @property
    def is_converter_logger(self):
        return self.__is_converter_logger

    @is_converter_logger.setter
    def is_converter_logger(self, value):
        self.__is_converter_logger = value

    def reset(self):
        with TbLogger.ERRORS_MUTEX:
            TbLogger.ALL_ERRORS_COUNT = max(0, TbLogger.ALL_ERRORS_COUNT - self.errors)
        self.errors = 0
        self._update_errors_batch()

    def stop(self):
        with TbLogger.ERRORS_MUTEX:
            TbLogger.ERRORS_BATCH.pop(self.attr_name, None)
        self.reset()
        from thingsboard_gateway.tb_utility.tb_handler import TBRemoteLoggerHandler
        for handler in self.handlers:
            if isinstance(handler, TBRemoteLoggerHandler):
                handler.remove_logger(self.name)
        self.handlers.clear()

    def _send_errors(self):
        with TbLogger.ERRORS_MUTEX:
            TbLogger.ERRORS_BATCH[self.attr_name] = 0
        self._is_on_init_state = False

    def trace(self, msg, *args, **kwargs):
        if self.isEnabledFor(TRACE_LOGGING_LEVEL):
            self._log(TRACE_LOGGING_LEVEL, msg, args, **kwargs)

    def error(self, msg, *args, **kwargs):
        attr_name = kwargs.pop('attr_name', None)
        kwargs['stacklevel'] = 2
        super(TbLogger, self).error(msg, *args, **kwargs)

        if self.__is_connector_logger:
            StatisticsService.count_connector_message(self.name, 'connectorsErrors')
        if self.__is_converter_logger:
            StatisticsService.count_connector_message(self.name, 'convertersErrors')

        self._add_error()
        self._update_errors_batch(error_attr_name=attr_name)

    def exception(self, msg, *args, **kwargs) -> None:
        attr_name = kwargs.pop('attr_name', None)
        kwargs['stacklevel'] = 2
        super(TbLogger, self).exception(msg, *args, **kwargs)

        if self.__is_connector_logger:
            StatisticsService.count_connector_message(self.name, 'connectorsErrors')
        if self.__is_converter_logger:
            StatisticsService.count_connector_message(self.name, 'convertersErrors')

        self._add_error()
        self._update_errors_batch(error_attr_name=attr_name)

    def _add_error(self):
        with TbLogger.ERRORS_MUTEX:
            TbLogger.ALL_ERRORS_COUNT += 1
            if TbLogger.PREVIOUS_ERRORS_RESET_TIME != self.__previous_reset_errors_time:
                self.__previous_reset_errors_time = TbLogger.PREVIOUS_ERRORS_RESET_TIME
                self.errors = 0
        self.errors += 1

    def _update_errors_batch(self, error_attr_name=None):
        if error_attr_name:
            error_attr_name = error_attr_name + '_ERRORS_COUNT'
        else:
            error_attr_name = self.attr_name
        TbLogger.ERRORS_BATCH[error_attr_name] = max(0, self.errors)

    @classmethod
    def send_errors_if_needed(cls, gateway):
        current_monotonic = monotonic()
        if (gateway.tb_client is not None
                and gateway.tb_client.is_connected()
                and (current_monotonic - cls.PREVIOUS_ERRORS_SENT_TIME >= cls.SEND_ERRORS_PERIOD
                     or cls.PREVIOUS_ERRORS_SENT_TIME == 0)):
            if (current_monotonic - cls.PREVIOUS_ERRORS_RESET_TIME >= cls.RESET_ERRORS_PERIOD
                    or cls.PREVIOUS_ERRORS_RESET_TIME == 0):
                cls.PREVIOUS_ERRORS_RESET_TIME = current_monotonic
                with cls.ERRORS_MUTEX:
                    for key in cls.ERRORS_BATCH:
                        cls.ERRORS_BATCH[key] = 0
                    cls.ALL_ERRORS_COUNT = 0
            cls.PREVIOUS_ERRORS_SENT_TIME = current_monotonic
            batch_to_send = {}
            with cls.ERRORS_MUTEX:
                if cls.ERRORS_BATCH:
                    batch_to_send = cls.ERRORS_BATCH
                if cls.__PREVIOUS_ALL_ERRORS_COUNT != cls.ALL_ERRORS_COUNT:
                    batch_to_send['ALL_ERRORS_COUNT'] = cls.ALL_ERRORS_COUNT
                    cls.__PREVIOUS_ALL_ERRORS_COUNT = cls.ALL_ERRORS_COUNT
            if batch_to_send:
                previous_batch_keys = set(cls.__PREVIOUS_BATCH_TO_SEND.keys())
                batch_keys = set(batch_to_send.keys())
                keys_to_check = set(previous_batch_keys).union(batch_keys)
                batch_sending_required = False
                for key in keys_to_check:
                    if cls.__PREVIOUS_BATCH_TO_SEND.get(key) != batch_to_send.get(key):
                        batch_sending_required = True
                        break
                if batch_sending_required:
                    gateway.send_telemetry(batch_to_send)
                    with cls.ERRORS_MUTEX:
                        keys_to_remove = previous_batch_keys - batch_keys
                        for key in keys_to_remove:
                            cls.__PREVIOUS_BATCH_TO_SEND.pop(key, None)
                        cls.__PREVIOUS_BATCH_TO_SEND.update(batch_to_send)

    @staticmethod
    def is_main_module_logger(name, attr_name, is_converter_logger):
        return name == attr_name or attr_name is None or (name != attr_name and is_converter_logger)

    @staticmethod
    def update_file_handlers():
        for logger in logging.Logger.manager.loggerDict.values():
            if hasattr(logger, 'is_connector_logger') or hasattr(logger, 'is_converter_logger'):
                file_handler_filter = list(filter(lambda handler: isinstance(handler, TimedRotatingFileHandler),
                                                  logger.handlers))
                if len(file_handler_filter):
                    old_file_handler = file_handler_filter[0]

                    new_file_handler = None
                    filename = old_file_handler.baseFilename.split(path.sep)[-1].split('.')[0]
                    if logger.is_connector_logger:
                        new_file_handler = TimedRotatingFileHandler.get_connector_file_handler(filename) # noqa

                    if logger.is_converter_logger:
                        new_file_handler = TimedRotatingFileHandler.get_converter_file_handler(filename) # noqa

                    if new_file_handler:
                        logger.addHandler(new_file_handler)
                        logger.removeHandler(old_file_handler)

    @staticmethod
    def check_and_update_file_handlers_class_name(config):
        for handler_config in config.get('handlers', {}).values():
            if handler_config.get('class', '') == 'thingsboard_gateway.tb_utility.tb_handler.TimedRotatingFileHandler':
                handler_config['class'] = 'thingsboard_gateway.tb_utility.tb_rotating_file_handler.TimedRotatingFileHandler' # noqa


logging.setLoggerClass(TbLogger)



from thingsboard_gateway.connectors.modbus.constants import PymodbusDefaults

from pymodbus.client.tls import AsyncModbusTlsClient as TlsClient
from pymodbus.framer.base import FramerType


class AsyncModbusTlsClient(TlsClient):
    def __init__(
        self,
        host: str,
        port: int = PymodbusDefaults.TlsPort,
        framer: FramerType = FramerType.TLS,
        certfile: str = None,
        keyfile: str = None,
        password: str = None,
        **kwargs: any,
    ):
        sslctx = AsyncModbusTlsClient.generate_ssl(certfile=certfile, keyfile=keyfile, password=password)
        retries = kwargs.pop('retries', PymodbusDefaults.Retries)
        timeout = kwargs.pop('timeout', PymodbusDefaults.Timeout)
        super().__init__(host=host, port=port, framer=framer, sslctx=sslctx, retries=retries, timeout=timeout)



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from typing import List, Union

from pymodbus.constants import Endian
from pymodbus.exceptions import ModbusIOException
from pymodbus.payload import BinaryPayloadDecoder
from pymodbus.pdu import ExceptionResponse

from thingsboard_gateway.connectors.modbus.entities.bytes_uplink_converter_config import BytesUplinkConverterConfig
from thingsboard_gateway.connectors.modbus.modbus_converter import ModbusConverter
from thingsboard_gateway.gateway.entities.converted_data import ConvertedData
from thingsboard_gateway.gateway.entities.report_strategy_config import ReportStrategyConfig
from thingsboard_gateway.gateway.statistics.decorators import CollectStatistics
from thingsboard_gateway.gateway.statistics.statistics_service import StatisticsService
from thingsboard_gateway.tb_utility.tb_utility import TBUtility


class BytesModbusUplinkConverter(ModbusConverter):
    def __init__(self, config: BytesUplinkConverterConfig, logger):
        self._log = logger
        self.__config = config

    @CollectStatistics(start_stat_type='receivedBytesFromDevices',
                       end_stat_type='convertedBytesFromDevice')
    def convert(self, _, data: List[dict]) -> Union[ConvertedData, None]:
        result = ConvertedData(self.__config.device_name, self.__config.device_type)
        device_report_strategy = self._get_device_report_strategy(self.__config.report_strategy,
                                                                  self.__config.device_name)

        converted_data_append_methods = {
            'attributes': result.add_to_attributes,
            'telemetry': result.add_to_telemetry
        }
        for device_data in data:
            StatisticsService.count_connector_message(self._log.name, 'convertersMsgProcessed')

            for config_section in converted_data_append_methods:
                for config in getattr(self.__config, config_section):
                    encoded_data = device_data[config_section].get(config['tag'])

                    if encoded_data:
                        decoded_data = self.decode_data(encoded_data, config,
                                                        self.__config.byte_order,
                                                        self.__config.word_order)

                        if decoded_data is not None:
                            datapoint_key = TBUtility.convert_key_to_datapoint_key(config['tag'],
                                                                                   device_report_strategy,
                                                                                   config,
                                                                                   self._log)
                            converted_data_append_methods[config_section]({datapoint_key: decoded_data})

            self._log.trace("Decoded data: %s", result)
            StatisticsService.count_connector_message(self._log.name, 'convertersAttrProduced',
                                                      count=result.attributes_datapoints_count)
            StatisticsService.count_connector_message(self._log.name, 'convertersTsProduced',
                                                      count=result.telemetry_datapoints_count)

            return result

    def decode_data(self, encoded_data, config, endian_order, word_endian_order):
        decoded_data = None

        if not isinstance(encoded_data, ModbusIOException) and not isinstance(encoded_data, ExceptionResponse):
            if config['functionCode'] in (1, 2):
                try:
                    decoder = self.from_coils(encoded_data.bits, endian_order=endian_order,
                                              word_endian_order=word_endian_order)
                except TypeError:
                    decoder = self.from_coils(encoded_data.bits, word_endian_order=word_endian_order)

                decoded_data = self.decode_from_registers(decoder, config)
            elif config['functionCode'] in (3, 4):
                decoder = BinaryPayloadDecoder.fromRegisters(encoded_data.registers, byteorder=endian_order,
                                                             wordorder=word_endian_order)
                decoded_data = self.decode_from_registers(decoder, config)

                if config.get('divider'):
                    decoded_data = float(decoded_data) / float(config['divider'])
                elif config.get('multiplier'):
                    decoded_data = decoded_data * config['multiplier']
        else:
            self._log.exception("Error while decoding data: %s, with config: %s", encoded_data, config)
            decoded_data = None

        return decoded_data

    @staticmethod
    def from_coils(coils, endian_order=Endian.LITTLE, word_endian_order=Endian.BIG):
        _is_wordorder = '_wordorder' in BinaryPayloadDecoder.fromCoils.__code__.co_varnames
        if _is_wordorder:
            try:
                decoder = BinaryPayloadDecoder.fromCoils(coils, byteorder=endian_order,
                                                         _wordorder=word_endian_order)
            except TypeError:
                decoder = BinaryPayloadDecoder.fromCoils(coils, _wordorder=word_endian_order)
        else:
            try:
                decoder = BinaryPayloadDecoder.fromCoils(coils, byteorder=endian_order,
                                                         wordorder=word_endian_order)
            except TypeError:
                decoder = BinaryPayloadDecoder.fromCoils(coils, wordorder=word_endian_order)

        return decoder

    def decode_from_registers(self, decoder, configuration):
        objects_count = configuration.get("objectsCount",
                                          configuration.get("registersCount", configuration.get("registerCount", 1)))
        lower_type = configuration["type"].lower()

        decoder_functions = {
            'string': decoder.decode_string,
            'bytes': decoder.decode_string,
            'bit': decoder.decode_bits,
            'bits': decoder.decode_bits,
            '8int': decoder.decode_8bit_int,
            '8uint': decoder.decode_8bit_uint,
            '16int': decoder.decode_16bit_int,
            '16uint': decoder.decode_16bit_uint,
            '16float': decoder.decode_16bit_float,
            '32int': decoder.decode_32bit_int,
            '32uint': decoder.decode_32bit_uint,
            '32float': decoder.decode_32bit_float,
            '64int': decoder.decode_64bit_int,
            '64uint': decoder.decode_64bit_uint,
            '64float': decoder.decode_64bit_float,
            }

        decoded = None

        if lower_type in ['bit', 'bits']:
            decoded = decoder_functions[lower_type]()
            decoded_lastbyte = decoder_functions[lower_type]()
            decoded += decoded_lastbyte
            decoded = decoded[len(decoded)-objects_count:]

        elif lower_type == "string":
            decoded = decoder_functions[lower_type](objects_count * 2)

        elif lower_type == "bytes":
            decoded = decoder_functions[lower_type](size=objects_count * 2)

        elif decoder_functions.get(lower_type) is not None:
            decoded = decoder_functions[lower_type]()

        elif lower_type in ['int', 'long', 'integer']:
            type_ = str(objects_count * 16) + "int"
            assert decoder_functions.get(type_) is not None
            decoded = decoder_functions[type_]()

        elif lower_type in ["double", "float"]:
            type_ = str(objects_count * 16) + "float"
            assert decoder_functions.get(type_) is not None
            decoded = decoder_functions[type_]()

        elif lower_type == 'uint':
            type_ = str(objects_count * 16) + "uint"
            assert decoder_functions.get(type_) is not None
            decoded = decoder_functions[type_]()

        else:
            self._log.error("Unknown type: %s", lower_type)

        if isinstance(decoded, int):
            result_data = decoded
        elif isinstance(decoded, bytes) and lower_type == "string":
            try:
                result_data = decoded.decode('UTF-8')
            except UnicodeDecodeError as e:
                self._log.error("Error decoding string from bytes, will be saved as hex: %s", decoded, exc_info=e)
                result_data = decoded.hex()
        elif isinstance(decoded, bytes) and lower_type == "bytes":
            result_data = decoded.hex()
        elif isinstance(decoded, list):
            if configuration.get('bit') is not None:
                result_data = int(decoded[configuration['bit'] if
                                          configuration['bit'] < len(decoded) else len(decoded) - 1])
            else:
                bitAsBoolean = configuration.get('bitTargetType', 'bool') == 'bool'
                if objects_count == 1:
                    result_data = bool(decoded[-1]) if bitAsBoolean else int(decoded[-1])
                else:
                    result_data = [bool(bit) if bitAsBoolean else int(bit) for bit in decoded]
        elif isinstance(decoded, float):
            result_data = float(round(decoded, configuration.get('round', 6)))
        elif decoded is not None:
            result_data = int(decoded, 16)
        else:
            result_data = decoded

        return result_data

    def _get_device_report_strategy(self, report_strategy, device_name):
        try:
            return ReportStrategyConfig(report_strategy)
        except ValueError as e:
            self._log.trace("Report strategy config is not specified for device %s: %s", device_name, e)



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from re import match, compile
from ipaddress import IPv4Network
import asyncio
from asyncio import Queue, CancelledError, QueueEmpty
from copy import deepcopy
from datetime import time
from threading import Thread
from string import ascii_lowercase
from random import choice
from time import monotonic, sleep
from typing import TYPE_CHECKING, Tuple, Any
from concurrent.futures import TimeoutError

from thingsboard_gateway.connectors.bacnet.constants import SUPPORTED_OBJECTS_TYPES, ALLOWED_APDU
from typing import TYPE_CHECKING
from ast import literal_eval

from thingsboard_gateway.connectors.bacnet.ede_parser import EDEParser
from thingsboard_gateway.connectors.bacnet.entities.routers import Routers
from thingsboard_gateway.connectors.connector import Connector
from thingsboard_gateway.gateway.constants import STATISTIC_MESSAGE_RECEIVED_PARAMETER, \
    STATISTIC_MESSAGE_SENT_PARAMETER, ON_ATTRIBUTE_UPDATE_DEFAULT_TIMEOUT, RPC_DEFAULT_TIMEOUT
from thingsboard_gateway.gateway.statistics.statistics_service import StatisticsService
from thingsboard_gateway.tb_utility.tb_logger import init_logger
from thingsboard_gateway.tb_utility.tb_utility import TBUtility

try:
    from bacpypes3.apdu import ErrorRejectAbortNack
except ImportError:
    print("bacpypes3 library not found")
    TBUtility.install_package("bacpypes3")
    from bacpypes3.apdu import ErrorRejectAbortNack

from bacpypes3.pdu import Address, IPv4Address
from bacpypes3.primitivedata import Null, Real
from bacpypes3.basetypes import DailySchedule, TimeValue, DeviceObjectPropertyReference
from thingsboard_gateway.connectors.bacnet.device import Device, Devices
from thingsboard_gateway.connectors.bacnet.entities.device_object_config import DeviceObjectConfig
from thingsboard_gateway.connectors.bacnet.application import Application
from thingsboard_gateway.connectors.bacnet.backward_compatibility_adapter import BackwardCompatibilityAdapter

if TYPE_CHECKING:
    from thingsboard_gateway.gateway.tb_gateway_service import TBGatewayService


class AsyncBACnetConnector(Thread, Connector):
    def __init__(self, gateway, config, connector_type):
        self.statistics = {STATISTIC_MESSAGE_RECEIVED_PARAMETER: 0,
                           STATISTIC_MESSAGE_SENT_PARAMETER: 0}
        self.__connector_type = connector_type
        super().__init__()
        self.__gateway: 'TBGatewayService' = gateway
        self.__config = config
        self.name = config.get('name', 'BACnet ' + ''.join(choice(ascii_lowercase) for _ in range(5)))
        remote_logging = self.__config.get('enableRemoteLogging', False)
        log_level = self.__config.get('logLevel', "INFO")

        self.__log = init_logger(self.__gateway, self.name, log_level,
                                 enable_remote_logging=remote_logging,
                                 is_connector_logger=True)
        self.__converter_log = init_logger(self.__gateway, self.name + '_converter', log_level,
                                           enable_remote_logging=remote_logging,
                                           is_converter_logger=True, attr_name=self.name)
        self.__log.info('Starting BACnet connector...')

        if EDEParser.is_ede_config(self.__config):
            self.__log.info('EDE config detected, parsing...')
            self.__parse_ede_config()
            self.__log.debug('EDE config parsed')

        if BackwardCompatibilityAdapter.is_old_config(config):
            backward_compatibility_adapter = BackwardCompatibilityAdapter(config, self.__log)
            self.__config = backward_compatibility_adapter.convert()

        self.__id = self.__config.get('id')
        self.daemon = True
        self.__stopped = False
        self.__connected = False

        self.__application = None

        try:
            self.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.loop)
        except RuntimeError:
            self.loop = asyncio.get_event_loop()

        self.__process_device_queue = Queue(1_000_000)
        self.__process_device_rescan_queue = Queue(1_000_000)
        self.__data_to_convert_queue = Queue(1_000_000)
        self.__data_to_save_queue = Queue(1_000_000)
        self.__indication_queue = Queue(1_000_000)

        self.loop.set_exception_handler(self.exception_handler)

        self.__update_devices_data_config()
        self.__devices = Devices()
        self.__routers_cache = Routers()
        self.__devices_discover_period = self.__config.get('devicesDiscoverPeriodSeconds', 30)
        self.__previous_discover_time = 0
        self.__devices_rescan_objects_period = self.__config['application'].get('devicesRescanObjectsPeriodSeconds', 60)

    def __parse_ede_config(self):
        try:
            parsed_ede_config = EDEParser.parse(self.__config)
            self.__config = parsed_ede_config
        except Exception as e:
            self.__log.error(f"Error parsing EDE config: {e}")

    def get_device_shared_attributes_keys(self, device_name):
        task = self.loop.create_task(self.__devices.get_device_by_name(device_name))
        started = monotonic()
        while not task.done() and not self.__stopped and monotonic() - started < 1.0:
            sleep(.02)

        try:
            device = task.result() if not task.cancelled() else None
            if device is not None:
                return device.shared_attributes_keys
        except Exception:
            return []

        return []

    def __update_devices_data_config(self):
        for device_config in self.__config.get('devices', []):
            new_attributes = []
            new_timeseries = []

            for section in ('attributes', 'timeseries'):
                if device_config.get(section, '') == '*':
                    if section == 'attributes':
                        new_attributes = '*'
                    else:
                        new_timeseries = '*'

                    continue

                for config_item in device_config.get(section, []):
                    try:
                        if ((isinstance(config_item['objectId'], str) and not config_item['objectId'] == '*') or isinstance(config_item['objectId'], list)) \
                                and isinstance(config_item['objectType'], list):
                            self.__log.warning('Invalid using list of object types with string (except "*") or list objectId in config item %s. Skipping...', config_item)  # noqa
                            continue

                        Device.parse_config_key(config_item)

                        ranges = Device.parse_ranges(config_item['objectId'])
                        if len(ranges) > 0:
                            for rng in ranges:
                                if section == 'attributes':
                                    new_attributes.extend([{**config_item, 'objectId': i} for i in range(*rng)])
                                else:
                                    new_timeseries.extend([{**config_item, 'objectId': i} for i in range(*rng)])
                        else:
                            if section == 'attributes':
                                new_attributes.append(config_item)
                            else:
                                new_timeseries.append(config_item)
                    except Exception as e:
                        self.__log.warning('Error parsing config item %s: %s', config_item, e)

            device_config['attributes'] = new_attributes
            device_config['timeseries'] = new_timeseries

    def exception_handler(self, _, context):
        if context.get('exception') is not None:
            if str(context['exception']) == 'invalid state transition from COMPLETED to AWAIT_CONFIRMATION' \
                    or str(context['exception']) == 'no more packet data':
                pass
            else:
                self.__log.exception('handled exception', exc_info=context['exception'])

    def open(self):
        self.start()

    def run(self):
        self.__connected = True

        try:
            self.loop.run_until_complete(self.__start())
        except CancelledError as e:
            self.__log.debug('Task was cancelled due to connector stop: %s', e.__str__())
        except Exception as e:
            self.__log.exception(e)

    async def __rescan_devices(self):
        while not self.__stopped:
            try:
                device = self.__process_device_rescan_queue.get_nowait()
                self.loop.create_task(self.__rescan(device))
            except QueueEmpty:
                await asyncio.sleep(.1)

    async def __start(self):
        if not self.__is_valid_application_device_section():
            self.__log.error('Can not start connector due to invalid application section in config.')
            return

        if self.__config.get('foreignDevice', {}).get('address', ''):
            self.__application = Application(DeviceObjectConfig(
                self.__config['application']), self.__handle_indication, self.__log,
                is_foreign_application=True)

            foreign_device_address = IPv4Address(self.__config['foreignDevice']['address'])
            foreign_device_ttl = int(self.__config['foreignDevice']['ttl'])
            self.__application.register_foreign_device(foreign_device_address, foreign_device_ttl)
        else:
            self.__application = Application(DeviceObjectConfig(
                self.__config['application']), self.__handle_indication, self.__log)

        await self.__discover_devices()
        await asyncio.gather(self.__main_loop(),
                             self.__rescan_devices(),
                             self.__convert_data(),
                             self.__save_data(),
                             self.indication_callback(),
                             self.__application.confirmation_handler())

    def __is_valid_application_device_section(self) -> bool:
        app = self.__config.get('application')
        if not isinstance(app, dict):
            self.__log.error("Missing or invalid 'application' section in config.")
            return False

        host = app.get('host')
        if not host:
            self.__log.error("Missing IPv4 address: 'application.address' (or 'application.host').")
            return False
        try:
            IPv4Address(str(host))
        except Exception:
            self.__log.error("Invalid IPv4 address in application section: %s", host)
            return False

        port = app.get('port')
        if not (1 <= int(port) <= 65535):
            self.__log.error(
                "The port inside application section must be in range [1, 65535], but got %d.",
                port
            )
            return False

        apdu = app.get('maxApduLengthAccepted', 1476)
        if apdu not in ALLOWED_APDU:
            self.__log.warning(
                "Unsupported value for 'maxApduLengthAccepted': %d. Allowed values are %s. Using default - 1476.",
                apdu, ALLOWED_APDU
            )
            app['maxApduLengthAccepted'] = 1476

        mask = app.get('mask', "24")
        try:
            if mask.isdigit() and not (0 <= int(mask) <= 32):
                self.__log.warning(
                    "The mask inside application section must be in range [0, 32], but got %s., using default - 24",
                    mask)
                app['mask'] = "24"
                return True
            else:
                IPv4Network(f"{host}/{mask}")
                return True
        except Exception:
            app['mask'] = "24"
            self.__log.warning("Invalid subnet mask inside application section : %s using default - 24", mask)

        return True

    def __handle_indication(self, apdu):
        self.__indication_queue.put_nowait(apdu)

    async def indication_callback(self):
        """
        First check if device is already added
        If added, set active to True
        If not added, check if device is in config
        """

        while not self.__stopped:
            try:
                apdu = self.__indication_queue.get_nowait()

                device_address = apdu.pduSource.__str__()
                self.__log.info('Received APDU, from %s, trying to find device...', device_address)

                # Continue if APDU has no iAmDeviceIdentifier or it is None
                if not hasattr(apdu, 'iAmDeviceIdentifier') or apdu.iAmDeviceIdentifier is None:
                    continue

                added_device = await self.__devices.get_device_by_id(apdu.iAmDeviceIdentifier[1])
                if added_device is None:
                    self.__log.debug('Device %s not found in devices list', device_address)
                    device_config = Device.find_self_in_config(self.__config['devices'], apdu)
                    if device_config:
                        self.__log.debug('Device %s found in config. Adding...', device_address)
                        self.loop.create_task(self.__add_device(apdu, device_config))
                    else:
                        self.__log.debug('Device %s not found in config', device_address)
                else:
                    added_device.active = True
                    self.__log.debug('Device %s already added', added_device)
            except QueueEmpty:
                await asyncio.sleep(.1)
            except Exception as e:
                self.__log.error('Error processing indication callback: %s', e)

    async def __add_device(self, apdu, device_config):
        await self.__set_additional_device_info_to_apdu(apdu, device_config)

        device_config['devicesRescanObjectsPeriodSeconds'] = self.__devices_rescan_objects_period
        device = Device(self.connector_type,
                        device_config,
                        apdu,
                        self.__process_device_queue,
                        self.__process_device_rescan_queue,
                        self.__log,
                        self.__converter_log)

        await self.__devices.add(device)
        self.__gateway.add_device(device.device_info.device_name,
                                  {"connector": self},
                                  device_type=device.device_info.device_type)
        self.__log.info('Device %s connected to platform', device.device_info.device_name)

        self.loop.create_task(device.run())
        self.__log.debug('Device %s started', device)

        self.__log.debug('Checking device %s configuration...', device.device_info.device_name)
        await self.__check_and_update_device_config(device)
        self.__log.debug('Checked device %s configuration.', device.device_info.device_name)

        self.loop.create_task(device.rescan())

    async def __set_additional_device_info_to_apdu(self, apdu, device_config):
        if Device.need_to_retrieve_device_name(device_config):
            apdu.deviceName = None
            device_name = await self.__application.get_device_name(apdu.pduSource, apdu.iAmDeviceIdentifier)

            if device_name is not None:
                apdu.deviceName = device_name

        if Device.need_to_retrieve_router_info(device_config):
            try:
                address_net = apdu.pduSource.addrNet

                router_info = await self.__routers_cache.get_router_info_by_address(address_net)
                if router_info is None:
                    router_info = await self.__get_router_info(address_net)
                    if router_info is not None:
                        await self.__routers_cache.add_router_info(address_net, router_info)
                    else:
                        self.__log.warning('Router not found for device %s', device_config['address'])

                if router_info is not None:
                    apdu.routerName = router_info['routerName']
                    apdu.routerAddress = router_info['routerAddress']
                    apdu.routerId = router_info['routerId']
                    apdu.routerVendorId = router_info['routerVendorId']
                    self.__log.debug('Router info for device %s: %s', device_config['address'], router_info)
                else:
                    self.__log.error('Failed to retrieve router %s info', device_config['address'])
            except Exception as e:
                self.__log.error('Error retrieving router info for device %s: %s', device_config['address'], e)

    async def __get_router_info(self, address_net):
        result = list(filter(lambda net: address_net in net[1],
                             self.__application.elementService.clientPeer.router_info_cache.router_dnets.items()))

        if len(result) > 0:
            (_, router_address), _ = result[0]

            router_info = await self.__application.get_router_info(router_address)
            return router_info

    async def __rescan(self, device):
        await self.__local_objects_discovery(device,
                                             device.rescan_objects_config,
                                             index_to_read=device.details.failed_to_read_indexes)

    async def __check_and_update_device_config(self, device, index_to_read=None):
        discover_for = Device.is_global_discovery_config(device.config)
        if len(discover_for):
            await self.__global_objects_discovery(device, discover_for, index_to_read=index_to_read)

        discover_for = Device.is_local_discovery_config(device.config)
        if len(discover_for):
            await self.__local_objects_discovery(device, discover_for, index_to_read=index_to_read)

    async def __global_objects_discovery(self, device, discover_for, index_to_read=None):
        self.__log.info('Discovering %s device objects...', device.details.address)

        new_config = deepcopy(device.config)
        discovering_started = monotonic()

        iter = await self.__application.get_device_objects(device, index_to_read=index_to_read)
        if iter is not None:
            all_done = False
            while not self.__stopped and not all_done:
                objects, _, all_done = await iter.get_next()
                config = self.from_objects_to_config(device, objects)

                for section in discover_for:
                    if isinstance(new_config[section], str):
                        new_config[section] = []
                        device.rescan_objects_config.append({
                            'type': section,
                            'objectType': '*',
                            'objectId': '*',
                            'propertyId': {'presentValue', 'objectName'},
                            'key': '${objectName}'
                        })

                    new_config[section].extend(config)

                device.config = new_config

        self.__log.debug('Device %s objects discovered, discovering took (sec): %r',
                         device.details.address,
                         monotonic() - discovering_started)

    def from_objects_to_config(self, device, object_list):
        config = []
        for obj in object_list:
            try:
                obj_id, _, _, key = obj
                config.append({
                    'objectType': str(obj_id[0]),
                    'objectId': obj_id[-1],
                    'propertyId': 'presentValue',
                    'key': key
                })
            except Exception as e:
                self.__log.error('Error converting object to config for device %s: %s', device, e)
                continue

        return config

    async def __local_objects_discovery(self, device, discover_for, index_to_read=None):
        self.__log.info('Discovering %s device objects...', device.details.address)

        new_config = deepcopy(device.config)

        with_all_properties = any(item['propertyId'] == '*' for item in discover_for)
        iter = await self.__application.get_device_objects(device,
                                                           with_all_properties=with_all_properties,
                                                           index_to_read=index_to_read)
        if iter is not None:
            all_done = False
            items_to_remove = []
            while not self.__stopped and not all_done:
                objects, _, all_done = await iter.get_next()

                for item_config in discover_for:
                    current_obj_id = None

                    for obj in objects:
                        new_config_item = {}
                        obj_id, prop_id, _, _ = obj
                        obj_type = str(obj_id[0])

                        if current_obj_id is None or current_obj_id != obj_id[-1]:
                            current_obj_id = obj_id[-1]

                            camel_case_obj_type = TBUtility.kebab_case_to_camel_case(obj_type)
                            if camel_case_obj_type in item_config['objectType'] or item_config['objectType'] == '*':
                                new_config_item = {
                                    **item_config,
                                    'objectType': obj_type,
                                    'objectId': obj_id[-1],
                                }

                                if with_all_properties:
                                    if not isinstance(new_config_item['propertyId'], set):
                                        new_config_item['propertyId'] = set()

                                    new_config_item['propertyId'].add(str(prop_id))

                            else:
                                self.__log.debug('Object type %s not matching config %s for device %s',
                                                 obj_type, item_config['objectType'], device.details.address)
                        else:
                            new_config_item['propertyId'].add(str(prop_id))

                        if item_config['type'] == 'attributes':
                            new_config['attributes'].append(new_config_item)
                        elif item_config['type'] == 'timeseries':
                            new_config['timeseries'].append(new_config_item)

                        if item_config not in items_to_remove:
                            items_to_remove.append(item_config)

            for item in items_to_remove:
                if item in new_config[item['type']]:
                    new_config[item['type']].remove(item)
                device.rescan_objects_config.append(item)

            device.config = new_config

    async def __discover_devices(self):
        self.__previous_discover_time = monotonic()
        self.__log.info('Discovering devices...')
        for device_config in self.__config.get('devices', []):
            if self.__stopped:
                break

            try:
                DeviceObjectConfig.update_address_in_config_util(device_config)

                who_is_address = Device.get_who_is_address(device_config['address'])
                if who_is_address is None:
                    self.__log.error('Invalid address %s', device_config['address'])
                    continue

                await self.__application.do_who_is(device_address=who_is_address)
                self.__log.debug('WhoIs request sent to device %s', device_config['address'])
            except Exception as e:
                self.__log.error('Error discovering device %s: %s', device_config['address'], e)

    async def __main_loop(self):
        while not self.__stopped:
            try:
                if monotonic() - self.__previous_discover_time >= self.__devices_discover_period:
                    await self.__discover_devices()
            except Exception as e:
                self.__log.error('Error in main loop during discovering devices: %s', e)
                await asyncio.sleep(1)

            try:
                device: Device = self.__process_device_queue.get_nowait()
                if device.stopped:
                    self.__log.trace('Device %s stopped', device)
                    continue

                self.__log.info('Reading data from device %s...', device.details.address)
                self.loop.create_task(self.__read_multiple_properties(device))
                # TODO: Add handling for device activity/inactivity
            except QueueEmpty:
                await asyncio.sleep(.1)
            except Exception as e:
                self.__log.error('Error processing device requests: %s', e)

    async def __read_multiple_properties(self, device):
        reading_started = monotonic()

        iter = await self.__application.get_device_values(device)
        if iter is not None:
            all_done = False
            while not self.__stopped and not all_done:
                results, config, all_done = await iter.get_next()
                if len(results) > 0:
                    self.__log.trace('%s reading results: %s', device, results)
                    self.__data_to_convert_queue.put_nowait((device, config, results))

        reading_ended = monotonic()
        current_reading_time = reading_ended - reading_started

        if current_reading_time > device.poll_period:
            device.poll_period = current_reading_time
        elif current_reading_time < device.poll_period:
            if current_reading_time < device.original_poll_period:
                device.poll_period = device.original_poll_period
            else:
                device.poll_period = current_reading_time

    async def __read_property(self, address, object_id, property_id):
        try:
            result = await self.__application.read_property(address, object_id, property_id)
            return {"value": str(result)}
        except ErrorRejectAbortNack as err:
            return {"error": str(err)}
        except Exception as e:
            result = {}
            self.__log.error('Error reading property %s:%s from device %s: %s', object_id, property_id, address, e)
            result['error'] = str(e)
            return result

    async def __write_property(self, address, object_id, property_id, value, priority=None):
        result = {}
        try:
            if value is None and priority is None:
                self.__log.error('Value and priority are both None for property %s:%s on device %s. Cannot write.',
                                 object_id, property_id, address)
                return ValueError('Value and priority cannot be None')

            if priority is not None:
                priority = int(priority)
                if priority < 1 or priority > 16:
                    self.__log.error('Invalid priority %s for property %s:%s on device %s. Using default priority 8.',
                                     priority, object_id, property_id, address)
                    return ValueError('Invalid priority value')

            if value is None:
                value = Null(())

            if property_id == "weeklySchedule":
                value = self.__prepare_weekly_schedule_value(value)

            if property_id == "listOfObjectPropertyReferences":
               value = self.__prepare_list_of_object_property_references_value(value, property_id)

            await self.__application.write_property(address, object_id, property_id, value, priority=priority)
            result['value'] = value
            return result

        except ErrorRejectAbortNack as err:
            self.__log.error("BACnet %s on write to device with object_id: %s, property_id: %s, address: %s", str(err),
                             object_id, property_id, address)
            return {'error': str(err)}

        except Exception as err:
            self.__log.error('Error writing property %s:%s to device %s: %s', object_id, property_id, address, err)
            return {'error': str(err)}

    async def __prepare_weekly_schedule_value(self, value):
        schedule = []
        raw_schedule = literal_eval(value)
        for idx, day in enumerate(raw_schedule):
            schedule.append(DailySchedule(daySchedule=[]))
            for sched in day:
                schedule[idx].daySchedule.append(
                    TimeValue(
                        time=time(int(sched[0].split(":")[0]), int(sched[0].split(":")[1])),
                        value=Real(sched[1])
                    )
                )
        return schedule

    async def __prepare_list_of_object_property_references_value(self, value, property_id):
        props = []
        raw_props = literal_eval(value)
        for prop in raw_props:
            props.append(
                DeviceObjectPropertyReference(
                    objectIdentifier = prop,
                    propertyIdentifier = property_id
                )
            )
        return props

    async def __convert_data(self):
        while not self.__stopped:
            try:
                device, config, values = self.__data_to_convert_queue.get_nowait()
                self.__log.trace('%s data to convert: %s', device, values)

                converted_data = device.uplink_converter.convert(config, values)
                self.__data_to_save_queue.put_nowait((device, converted_data))
            except QueueEmpty:
                await asyncio.sleep(.1)
            except Exception as e:
                self.__log.error('Error converting data: %s', e)

    async def __save_data(self):
        while not self.__stopped:
            try:
                device, data_to_save = self.__data_to_save_queue.get_nowait()
                self.__log.trace('%s data to save: %s', device, data_to_save)
                StatisticsService.count_connector_message(self.get_name(), stat_parameter_name='storageMsgPushed')
                self.__gateway.send_to_storage(self.get_name(), self.get_id(), data_to_save)
                self.statistics[STATISTIC_MESSAGE_SENT_PARAMETER] += 1
            except QueueEmpty:
                await asyncio.sleep(.1)
            except Exception as e:
                self.__log.error('Error saving data: %s', e)

    def close(self):
        self.__log.info('Stopping BACnet connector...')
        self.__connected = False
        self.__stopped = True

        self.__devices.stop_all()

        if self.__application:
            self.__application.close()

        asyncio.run_coroutine_threadsafe(self.__cancel_all_tasks(), self.loop)

        self.__check_is_alive()

        self.__log.info('BACnet connector stopped')
        self.__log.stop()

    def __check_is_alive(self):
        start_time = monotonic()

        while self.is_alive():
            if monotonic() - start_time > 10:
                self.__log.error("Failed to stop connector %s", self.get_name())
                break
            sleep(.1)

    async def __cancel_all_tasks(self):
        await asyncio.sleep(5)
        for task in asyncio.all_tasks(self.loop):
            task.cancel()

    def __create_task(self, func, args, kwargs):
        task = self.loop.create_task(func(*args, **kwargs))
        return task

    def on_attributes_update(self, content):
        try:
            self.__log.debug('Received Attribute Update request: %r', content)
            device = self.__get_device_by_name(payload=content)
            if device is None:
                self.__log.error('Device %s not found', content['device'])
                return
            if not device.attributes_updates:
                self.__log.error("No attribute mapping found for device %s", device.name)
                return
            data_section = content.get('data', {})

            attribute_update_config_list = [
                update_config
                for update_config in device.attributes_updates
                if update_config["key"] in content.get("data", {})
            ]
            self.__process_task_on_filtered_attribute_update_section(
                filtered_attribute_update_config=attribute_update_config_list, data_section=data_section, device=device)

        except Exception as e:
            self.__log.error('Error processing attribute update %s with error: %s', content, str(e))
            self.__log.debug("Error", exc_info=e)

    def __process_task_on_filtered_attribute_update_section(self, filtered_attribute_update_config: list,
                                                            data_section: dict, device: Device):
        for attribute_update_config in filtered_attribute_update_config:
            try:
                value = data_section.get(attribute_update_config['key'])
                object_id = Device.get_object_id(attribute_update_config)
                kwargs = {'priority': attribute_update_config.get('priority')}
                task = self.__create_task(self.__process_attribute_update,
                                          (Address(device.details.address),
                                           object_id,
                                           attribute_update_config['propertyId'],
                                           value),
                                          kwargs)
                task_completed, result = self.__wait_task_with_timeout(task=task,
                                                                       timeout=ON_ATTRIBUTE_UPDATE_DEFAULT_TIMEOUT,
                                                                       poll_interval=0.2)
                if not task_completed:
                    self.__log.error(
                        "Failed to process rpc request for key %s, timeout has been reached",
                        attribute_update_config['key'],
                    )
                    continue
                self.__log.debug("Successfully processed attribute update for key %s", attribute_update_config['key'])

            except TypeError as e:
                self.__log.error('%s for key %s ', str(e), attribute_update_config['objectId'])
                self.__log.debug("Error", exc_info=e)
                continue

            except ValueError as e:
                self.__log.error("Such number of object id is not supported %d for key %s",
                                 attribute_update_config['objectId'], attribute_update_config['key'])
                self.__log.debug("Error", exc_info=e)
                continue

            except Exception as e:
                self.__log.error("Could not process attribute update with error %s", str(e))
                self.__log.debug("Error", exc_info=e)
                continue

    def __get_device_by_name(self, payload: dict) -> Device | None:
        device_name = payload.get('device')
        if not device_name:
            self.__log.error('The attribute update request does not contain a device name %s', payload)

        try:
            task = self.loop.create_task(self.__devices.get_device_by_name(device_name))
            task_completed, device = self.__wait_task_with_timeout(task=task,
                                                                   timeout=10,
                                                                   poll_interval=0.2)
            if not task_completed:
                self.__log.debug("Failed to get device %s, the device look up task failed on timeout ", device_name)
                return None

            return device

        except Exception as e:
            self.__log.debug("Error getting device by name %s with error: %s", device_name, exc_info=e)

    @staticmethod
    def __wait_task_with_timeout(task: asyncio.Task, timeout: float, poll_interval: float = 0.2) -> Tuple[bool, Any]:
        start_time = monotonic()
        while not task.done():
            sleep(poll_interval)
            current_time = monotonic()
            if current_time - start_time >= timeout:
                task.cancel()
                return False, None
        return True, task.result()

    async def __process_attribute_update(self, address, object_id, property_id, value, priority=None):
        result = {}
        result['response'] = await self.__write_property(address, object_id, property_id, value, priority=priority)
        return result

    def server_side_rpc_handler(self, content):
        self.__log.debug('Received RPC request: %r', content)

        device = self.__get_device_by_name(payload=content)

        if not device:
            self.__log.error('Failed to found device %s', content.get('device'))
            self.__gateway.send_rpc_reply(device=content.get('device'),
                                          req_id=content['data'].get('id'),
                                          content={"result": {'error': "Device not found"}})
            return

        rpc_method_name = content.get('data', {}).get('method')
        if rpc_method_name is None:
            self.__log.error('Method name not found in RPC request: %r', content)
            self.__gateway.send_rpc_reply(device=device.device_info.device_name,
                                          req_id=content['data'].get('id'),
                                          content={"result": {'error': "No valid rpc method found"}})
            return

        if rpc_method_name not in ('get', 'set'):
            filtered_rpc_section_from_config = [rpc_config for rpc_config in device.server_side_rpc if
                                                rpc_config['method'] == rpc_method_name]
            if not filtered_rpc_section_from_config:
                self.__log.error("Neither of configured device rpc methods match with %s", rpc_method_name)
                self.__gateway.send_rpc_reply(
                    device=device.device_info.device_name,
                    req_id=content.get('data', {}).get('id'),
                    content={
                        "result": {"error": f"Neither of configured device rpc methods match with {rpc_method_name}"}}
                )
                return

            for rpc_config in filtered_rpc_section_from_config:

                err = self.__validate_device_rpc(method_rpc_config_section=rpc_config)
                if err:
                    self.__log.error(err["error"])
                    self.__gateway.send_rpc_reply(
                        device=device.device_info.device_name,
                        req_id=content.get('data', {}).get('id'),
                        content={"result": err}
                    )
                    return

                self.__process_rpc(rpc_method_name, rpc_config, content, device)
                self.__log.debug("Processed  device RPC request %s for device %s", rpc_method_name,
                                 device.device_info.device_name)
            return
        result = self.__check_and_process_reserved_rpc(rpc_method_name, device, content)
        return result

    @staticmethod
    def __validate_device_rpc(method_rpc_config_section: dict) -> dict | None:
        obj_type = method_rpc_config_section['objectType']
        req_type = method_rpc_config_section['requestType']

        if obj_type not in SUPPORTED_OBJECTS_TYPES.values():
            return {
                "error": (
                    f"Invalid objectType: '{obj_type}'. "
                    f"Expected one of {list(SUPPORTED_OBJECTS_TYPES.values())}."
                )
            }

        if req_type not in ('writeProperty', 'readProperty'):
            return {
                "error": (
                    f"Invalid requestType: '{req_type}'. "
                    f"Expected 'writeProperty' or 'readProperty'."
                )
            }

    def __process_rpc(self, rpc_method_name, rpc_config, content, device):
        try:
            object_id = Device.get_object_id(rpc_config)
            value = content.get('data', {}).get('params')

            kwargs = {'priority': rpc_config.get('priority'), 'value': value,
                      'request_type': rpc_config.get('requestType')}
            task = self.__create_task(self.__process_rpc_request,
                                      (Address(device.details.address),
                                       object_id,
                                       rpc_config['propertyId']),
                                      kwargs)
            task_completed, result = self.__wait_task_with_timeout(task=task,
                                                                   timeout=content.get("timeout", RPC_DEFAULT_TIMEOUT),
                                                                   poll_interval=0.2)
            if not task_completed:
                self.__log.error(
                    "Failed to process rpc request for %s, timeout has been reached",
                    device.name,
                )
                result = {"error": f"Timeout rpc has been reached for {device.name}"}

            elif task_completed:
                if result.get("response", {}).get("value", {}):
                    self.__log.info('Processed RPC request with result: %s', result)
                else:
                    self.__log.error(
                        "An error occurred during RPC request: %s",
                        result.get('response', {}).get('error', '')
                    )

            self.__gateway.send_rpc_reply(device=device.device_info.device_name,
                                          req_id=content['data'].get('id'),
                                          content={'result': str(result.get('response'))})
            return result

        except ValueError as e:
            self.__log.error('Value error processing RPC request %s: %s', rpc_method_name, e)
            self.__gateway.send_rpc_reply(device=device.device_info.device_name,
                                          req_id=content['data'].get('id'),
                                          content={"result": {'error': f'Could not find correct mapping for {str(e)}'}})

        except Exception as e:
            self.__log.error(
                'Error processing RPC request %s: %s', rpc_method_name, e)
            self.__gateway.send_rpc_reply(device=device.device_info.device_name,
                                          req_id=content['data'].get('id'),
                                          content={"result": {'error': str(e)}}, )

    async def __process_rpc_request(self, address, object_id, property_id, priority=None, value=None,
                                    request_type=None):
        if request_type == "readProperty" or (request_type is None and value is None and priority is None):
            response = await self.__read_property(address, object_id, property_id)
        else:
            response = await self.__write_property(address, object_id, property_id, value, priority=priority)
        return {"response": response}

    def __check_and_process_reserved_rpc(self, rpc_method_name, device, content):
        params_section = content['data'].get('params', {})
        if not params_section:
            self.__log.error('No params section found in RPC request: %r', content)
            self.__gateway.send_rpc_reply(device=device.device_info.device_name,
                                          req_id=content['data'].get('id'),
                                          content={"result": {"error": 'No params section found in RPC request'}})
            return

        get_pattern = compile(r'objectType=[A-Za-z]+;objectId=\d+;propertyId=[A-Za-z]+;')

        set_pattern = compile(r'objectType=[A-Za-z]+;objectId=\d+;propertyId=[A-Za-z]+;(priority=\d+;)?value=.+;')
        pattern = get_pattern if rpc_method_name == 'get' else set_pattern
        expected_schema = (
            "get objectType=<objectType>;objectId=<objectId>;propertyId=<propertyId>;" if rpc_method_name == 'get'
            else "set objectType=<objectType>;objectId=<objectId>;propertyId=<propertyId>;priority=<priority>;value=<value>;"
        )
        if not pattern.match(params_section):
            self.__log.error(f"The requested RPC does not match with the schema: {expected_schema}")
            self.__gateway.send_rpc_reply(device=device.device_info.device_name,
                                          req_id=content['data'].get('id'),
                                          content={"result": {
                                              "error": f"The requested RPC does not match with the schema: {expected_schema}"}})
            return
        params = {}
        for param in params_section.split(';'):
            try:
                (key, value) = param.split('=')
            except ValueError:
                continue

            if key and value:
                params[key] = value

        if params['objectType'] not in SUPPORTED_OBJECTS_TYPES.values():
            self.__gateway.send_rpc_reply(device=device.device_info.device_name,
                                          req_id=content['data'].get('id'),
                                          content={"result": {
                                              "error": f"The objectType must be from '{list(SUPPORTED_OBJECTS_TYPES.values())}, but got'{params['objectType']}"}})
            return

        if rpc_method_name == 'get':
            params['requestType'] = 'readProperty'
            content['data'].pop('params')
        elif rpc_method_name == 'set':
            params['requestType'] = 'writeProperty'
            content['data'].pop('params')
            content['data']['params'] = params.get('value')

        result = self.__process_rpc(rpc_method_name, params, content, device)
        return result

    def get_id(self):
        return self.__id

    def get_name(self):
        return self.name

    def get_type(self):
        return self.__connector_type

    @property
    def connector_type(self):
        return self.__connector_type

    def get_config(self):
        return self.__config

    def is_connected(self):
        return self.__connected

    def is_stopped(self):
        return self.__stopped



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import sys
from os import curdir, listdir, mkdir, path, environ

from thingsboard_gateway.gateway.constants import DEV_MODE_PARAMETER_NAME, TB_GW_DEV_DEBUG_SERVER_PORT
from thingsboard_gateway.gateway.tb_gateway_service import TBGatewayService
from thingsboard_gateway.gateway.hot_reloader import HotReloader
from thingsboard_gateway.tb_utility.tb_utility import TBUtility


def is_running_under_pycharm():
    return (
        "PYCHARM_HOSTED" in environ
        or "_pydevd_bundle" in sys.modules
    )

def main():
    is_dev_mode = TBUtility.str_to_bool(environ.get(DEV_MODE_PARAMETER_NAME, 'false'))

    if is_dev_mode and not is_running_under_pycharm:
        run_debug_server()

    if "logs" not in listdir(curdir):
        mkdir("logs")

    try:
        hot_reload = bool(sys.argv[1])
    except IndexError:
        hot_reload = False

    if hot_reload:
        HotReloader(TBGatewayService)
    else:
        config_path = __get_config_path(path.dirname(path.abspath(__file__)) + '/config/'.replace('/', path.sep))
        TBGatewayService(config_path + 'tb_gateway.json')


def daemon():
    config_path = __get_config_path("/etc/thingsboard-gateway/config/".replace('/', path.sep))
    TBGatewayService(config_path + "tb_gateway.json")


def __get_config_path(default_config_path):
    config_path = environ.get("TB_GW_CONFIG_DIR", default_config_path)
    if not config_path.endswith(path.sep):
        config_path += path.sep
    return config_path


def run_debug_server():
    try:
        import debugpy
    except ImportError:
        TBUtility.install_package("debugpy")
        import debugpy

    debugpy_port = int(environ.get("TB_GW_DEV_DEBUG_SERVER", TB_GW_DEV_DEBUG_SERVER_PORT))
    debugpy.listen(("0.0.0.0", debugpy_port))


if __name__ == '__main__':
    main()



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
from datetime import timedelta

from thingsboard_gateway.connectors.converter import Converter
from thingsboard_gateway.gateway.constants import REPORT_STRATEGY_PARAMETER
from thingsboard_gateway.gateway.entities.converted_data import ConvertedData
from thingsboard_gateway.gateway.entities.report_strategy_config import ReportStrategyConfig
from thingsboard_gateway.gateway.entities.telemetry_entry import TelemetryEntry
from thingsboard_gateway.gateway.statistics.statistics_service import StatisticsService
from thingsboard_gateway.gateway.statistics.decorators import CollectStatistics
from thingsboard_gateway.tb_utility.tb_utility import TBUtility


class SNMPUplinkConverter(Converter):
    def __init__(self, config, logger):
        self._log = logger
        self.__config = config

    @CollectStatistics(start_stat_type='receivedBytesFromDevices',
                       end_stat_type='convertedBytesFromDevice')
    def convert(self, config, data):
        device_name = self.__config['deviceName']
        device_type = self.__config['deviceType']

        converted_data = ConvertedData(device_name=device_name, device_type=device_type)

        device_report_strategy = None
        try:
            device_report_strategy = ReportStrategyConfig(self.__config.get(REPORT_STRATEGY_PARAMETER))
        except ValueError as e:
            self._log.trace("Report strategy config is not specified for device %s: %s", self.__config['deviceName'], e)

        try:
            for datatype in ('attributes', 'telemetry'):
                for datatype_config in config[datatype]:
                    data_key = datatype_config["key"]
                    item_data = data.get(data_key)
                    value = None
                    if isinstance(item_data, dict):
                        value = {str(k): str(v) for k, v in item_data.items()}
                    elif isinstance(item_data, list):
                        if isinstance(item_data[0], str):
                            value = ','.join(item_data)
                        elif isinstance(item_data[0], dict):
                            res = {}
                            for item in item_data:
                                res.update(**item)
                            value = {str(k): str(v) for k, v in res.items()}
                    elif isinstance(item_data, str):
                        value = item_data
                    elif isinstance(item_data, bytes):
                        value = item_data.decode("UTF-8")
                    elif isinstance(item_data, timedelta):
                        value = item_data.total_seconds()
                    else:
                        value = item_data

                    if value:
                        datapoint_key = TBUtility.convert_key_to_datapoint_key(data_key, device_report_strategy,
                                                                               datatype_config, self._log)
                        if datatype == 'attributes':
                            converted_data.add_to_attributes(datapoint_key, value)
                        else:
                            telemetry_entry = TelemetryEntry({datapoint_key: value})
                            converted_data.add_to_telemetry(telemetry_entry)
        except Exception as e:
            StatisticsService.count_connector_message(self._log.name, 'convertersMsgDropped')
            self._log.exception(e)

        self._log.debug(converted_data)
        StatisticsService.count_connector_message(self._log.name, 'convertersAttrProduced',
                                                  count=converted_data.attributes_datapoints_count)
        StatisticsService.count_connector_message(self._log.name, 'convertersTsProduced',
                                                  count=converted_data.telemetry_datapoints_count)

        return converted_data



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from asyncio import Lock


class Routers:
    def __init__(self):
        self.__lock = Lock()
        self.__router_info_cache = {}

    async def add_router_info(self, address, info):
        await self.__lock.acquire()
        try:
            if address not in self.__router_info_cache:
                self.__router_info_cache[address] = info
        finally:
            self.__lock.release()

    async def get_router_info_by_address(self, address):
        await self.__lock.acquire()
        try:
            return self.__router_info_cache.get(address)
        finally:
            self.__lock.release()



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from bacpypes3.apdu import IAmRequest
from bacpypes3.basetypes import Segmentation

MAX_REQUEST_SIZE = 26


class BACnetDeviceDetails:
    def __init__(self, i_am_request: IAmRequest):
        self.address = i_am_request.pduSource.__str__()
        self.__identifier = i_am_request.iAmDeviceIdentifier
        self.__object_identifier = i_am_request.iAmDeviceIdentifier[1]
        self.__vendor_id = i_am_request.vendorID
        self.__object_name = i_am_request.deviceName if hasattr(i_am_request, 'deviceName') else i_am_request.iAmDeviceIdentifier[1]  # noqa
        self.__max_apdu_length = i_am_request.maxAPDULengthAccepted
        self.__segmentation = i_am_request.segmentationSupported

        self.__router_id = i_am_request.routerId if hasattr(i_am_request, 'routerId') else None
        self.__router_name = i_am_request.routerName if hasattr(i_am_request, 'routerName') else None
        self.__router_address = i_am_request.routerAddress if hasattr(i_am_request, 'routerAddress') else None
        self.__router_vendor_id = i_am_request.routerVendorId if hasattr(i_am_request, 'routerVendorId') else None

        self.__objects_len = 0
        self.__failed_to_read_indexes = set()

    def __str__(self):
        return (f"DeviceDetails(address={self.address}, objectIdentifier={self.__object_identifier}, "
                f"vendorId={self.__vendor_id}, objectName={self.__object_name}")

    @property
    def object_id(self):
        return self.__object_identifier

    @property
    def identifier(self):
        return self.__identifier

    @property
    def max_apdu_length(self):
        return self.__max_apdu_length

    @property
    def vendor_id(self):
        return self.__vendor_id

    @property
    def as_dict(self):
        result = {
            "address": self.address,
            "objectId": self.__object_identifier,
            "vendorId": self.__vendor_id,
            "objectName": self.__object_name,
        }

        if self.__router_id is not None:
            result['routerId'] = self.__router_id
        if self.__router_name is not None:
            result['routerName'] = self.__router_name
        if self.__router_address is not None:
            result['routerAddress'] = self.__router_address
        if self.__router_vendor_id is not None:
            result['routerVendorId'] = self.__router_vendor_id

        return result

    @property
    def failed_to_read_indexes(self):
        return self.__failed_to_read_indexes

    @failed_to_read_indexes.setter
    def failed_to_read_indexes(self, value: int):
        if value not in self.__failed_to_read_indexes:
            self.__failed_to_read_indexes.add(value)

    @property
    def objects_len(self):
        return self.__objects_len

    @objects_len.setter
    def objects_len(self, value):
        self.__objects_len = value

    def sucess_read_for(self, value: int):
        if value in self.__failed_to_read_indexes:
            self.__failed_to_read_indexes.discard(value)

    def is_segmentation_supported(self):
        return self.__segmentation in (Segmentation.segmentedBoth, Segmentation.segmentedTransmit)

    def get_max_apdu_count(self) -> int:
        return int(self.__max_apdu_length / MAX_REQUEST_SIZE)



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from logging import getLogger, setLoggerClass
from platform import platform, release, system
from time import time
from uuid import uuid1

from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as metadata_version
from requests import ConnectionError, post
from simplejson import loads

from threading import Thread

from thingsboard_gateway import version
from thingsboard_gateway.tb_utility.tb_logger import TbLogger
from thingsboard_gateway.tb_utility.tb_utility import TBUtility

setLoggerClass(TbLogger)
log = getLogger("service")

UPDATE_SERVICE_BASE_URL = "https://updates.thingsboard.io"


class TBUpdater:
    def __init__(self):
        super().__init__()

        try:
            distribution_version = metadata_version('thingsboard_gateway')
            self.__version = {"current_version": distribution_version, "latest_version": distribution_version}
        except PackageNotFoundError:
            self.__version = {"current_version": version.VERSION, "latest_version": version.VERSION}

        self.__instance_id = str(uuid1())
        self.__platform = system()
        self.__release = release()
        self.__os_version = platform()
        self.__previous_check = 0
        self.__check_period = 3600.0
        self.__request_timeout = (1,2)
        self.__stopped = True
        Thread(target=self.check_for_new_version, daemon=True).start()

    def stop(self):
        self.__stopped = True

    def get_version(self):
        if time() >= self.__previous_check + self.__check_period:
            Thread(target=self.check_for_new_version, daemon=True).start()
            self.__previous_check = time()
        return self.__version

    def get_platform(self):
        return self.__platform

    def get_release(self):
        return self.__release

    def check_for_new_version(self):
        self.__previous_check = time()
        log.debug("Checking for new version")
        request_args = self.form_request_params()
        try:
            response = post(**request_args)
            content = None
            content = loads(response.content)
            if content is not None and content.get("updateAvailable", False):
                new_version = content["message"].replace("New version ", "").replace(" is available!", "")
                if new_version > self.__version["current_version"]:
                    log.info(content["message"])
                    self.__version["latest_version"] = new_version
                    log.info("\n\n[===UPDATE===]\n\n New version %s is available! \n\n[===UPDATE===]\n",
                             self.__version["latest_version"])
        except ConnectionRefusedError:
            log.warning("Cannot connect to the update service. Please check your internet connection.")
        except ConnectionError:
            log.warning("Cannot connect to the update service. Please check your internet connection.")
        except Exception as e:
            log.error("An error occurred while checking for a new version: %s", e)
            log.debug("Error details", exc_info=e)

    def form_request_params(self):
        json_data = {
            "version": self.__version["current_version"],
            "platform": self.__platform,
            "instanceId": self.__instance_id,
            "osVersion": self.__os_version,
            }
        url = UPDATE_SERVICE_BASE_URL + "/api/tb-gateway/updates"
        request_args = {
            "url": url,
            "json": json_data,
            "timeout": self.__request_timeout

            }
        return request_args

    def update(self):
        if self.__version["latest_version"] != self.__version["current_version"]:
            result = TBUtility.install_package("thingsboard-gateway", self.__version["latest_version"])
        else:
            result = "Congratulations! You have the latest version."
        return result


if __name__ == '__main__':
    updater = TBUpdater()



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.


from copy import deepcopy
import re
from typing import List

from thingsboard_gateway.tb_utility.tb_logger import TbLogger


class BackwardCompatibilityAdapter:
    DATA_TYPES = {
        'int': 'integer',
        'str': 'string',
        'float': 'double',
        'bool': 'boolean',
    }

    def __init__(self, config, logger: TbLogger):
        self._config = deepcopy(config)
        self._log = logger

    def convert(self):
        errors = self.validate_incoming_config(self._config)
        if errors:
            for error in errors:
                self._log.error("Found error in incoming configuration (Config will be ignored): %s", error)
        disable_subscriptions = self._config['server'].pop('disableSubscriptions', None)
        if disable_subscriptions is not None:
            self._config['server']['enableSubscriptions'] = not disable_subscriptions

        mapping_configuration = deepcopy(self._config.get('server', {}).get('mapping', []))
        if not mapping_configuration:
            return self._config

        for node_config in mapping_configuration:
            try:
                node_config['deviceNodeSource'] = self.get_value_source(node_config['deviceNodePattern'], False)

                device_type_pattern = node_config.pop('deviceTypePattern', 'default')
                device_name_pattern = node_config.pop('deviceNamePattern', None)
                node_config['deviceInfo'] = {
                    'deviceNameExpressionSource': self.get_value_source(device_name_pattern),
                    'deviceNameExpression': device_name_pattern,
                    'deviceProfileExpressionSource': self.get_value_source(device_type_pattern),
                    'deviceProfileExpression': device_type_pattern
                }

                mapping_datapoints_configurations = {}

                # converting attributes and timeseries sections
                for config_section_name in ('attributes', 'timeseries'):
                    mapping_datapoints_configurations[config_section_name] = []
                    for section_config in node_config.get(config_section_name, []):
                        try:
                            path = section_config.pop('path', None)

                            section_config['type'] = self.get_value_source(path)
                            section_config['value'] = path
                            mapping_datapoints_configurations[config_section_name].append(section_config)
                        except Exception as e:
                            pass

                    node_config[config_section_name] = mapping_datapoints_configurations[config_section_name]

                # converting attributes_updates section
                for config in node_config.get('attributes_updates', []):
                    attribute_on_tb = config.pop('attributeOnThingsBoard', None)
                    attribute_on_device = config.pop('attributeOnDevice', None)

                    config['key'] = attribute_on_tb
                    config['type'] = self.get_value_source(attribute_on_device)
                    config['value'] = attribute_on_device

                # converting rpc section
                for config in node_config.get('rpc_methods', []):
                    arguments = config.pop('arguments', [])
                    config['arguments'] = []

                    for arg in arguments:
                        converted_argument = {
                            'type': self.DATA_TYPES.get(type(arg).__name__, 'string'),
                            'value': arg
                        }
                        config['arguments'].append(converted_argument)
            except Exception as e:
                self._log.error('Error during conversion to new configuration format: %r', e)
                self._log.trace('Config section: %r', node_config)

        # Removing old mapping section
        self._config['server'].pop('mapping')
        # Adding new mapping section
        self._config['mapping'] = mapping_configuration

        return self._config

    @staticmethod
    def get_value_source(value, possible_constant=True):
        if re.search(r"(ns=\d+;[isgb]=[^}]+)", value):
            return 'identifier'
        elif re.search(r"\${([A-Za-z.:\\\d]+)}", value) or not possible_constant:
            return 'path'
        else:
            return 'constant'

    @staticmethod
    def validate_incoming_config(config) -> List[str]:
        errors = []
        if 'server' not in config:
            errors.append('No server configuration found')
        if 'mapping' not in config['server'] or not config['server']['mapping']:
            errors.append('No mapping configuration found')
        else:
            for index, node_config in enumerate(config['server']['mapping']):
                object_number = index + 1
                if 'deviceNodePattern' not in node_config:
                    errors.append('No "deviceNodePattern" found in %i object in mapping configuration' % object_number)
                if 'deviceNamePattern' not in node_config:
                    errors.append('No "deviceNamePattern" found in %i object in mapping configuration' % object_number)
                if 'attributes' not in node_config:
                    errors.append('No "attributes" found in %i object in mapping configuration' % object_number)
                else:
                    for attribute_number, attribute_config in enumerate(node_config['attributes']):
                        attribute_errors = BackwardCompatibilityAdapter.validate_datapoint_config(attribute_config)
                        if attribute_errors:
                            for error in attribute_errors:
                                errors.append('Attribute configuration with number %i in %i mapping object has the following error: %s' % (attribute_number + 1, object_number, error))

                if 'timeseries' not in node_config:
                    errors.append('No "timeseries" found in %i object in mapping configuration' % object_number)
                else:
                    for timeseries_number, timeseries_config in enumerate(node_config['timeseries']):
                        timeseries_errors = BackwardCompatibilityAdapter.validate_datapoint_config(timeseries_config)
                        if timeseries_errors:
                            for error in timeseries_errors:
                                errors.append('Timeseries configuration with number %i in %i mapping object has the following error: %s' % (timeseries_number + 1, object_number, error))
        return errors

    @staticmethod
    def validate_datapoint_config(config) -> List[str]:
        errors = []
        if 'path' not in config:
            errors.append('No "path" found in configuration')
        if 'key' not in config:
            errors.append('No "key" found in configuration')
        return errors



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from thingsboard_gateway.connectors.odbc.odbc_converter import OdbcConverter
from thingsboard_gateway.gateway.constants import TIMESERIES_PARAMETER, ATTRIBUTES_PARAMETER
from thingsboard_gateway.gateway.entities.converted_data import ConvertedData
from thingsboard_gateway.gateway.entities.report_strategy_config import ReportStrategyConfig
from thingsboard_gateway.gateway.entities.telemetry_entry import TelemetryEntry
from thingsboard_gateway.gateway.statistics.decorators import CollectStatistics
from thingsboard_gateway.gateway.statistics.statistics_service import StatisticsService
from thingsboard_gateway.tb_utility.tb_utility import TBUtility


class OdbcUplinkConverter(OdbcConverter):

    DATATYPES = (TIMESERIES_PARAMETER, ATTRIBUTES_PARAMETER)

    def __init__(self, logger):
        self._log = logger

    @CollectStatistics(start_stat_type='receivedBytesFromDevices',
                       end_stat_type='convertedBytesFromDevice')
    def convert(self, config, data):
        if isinstance(config, str) and config == "*":
            if "deviceName" in data and (('timeseries' in data or 'telemetry' in data) and 'attributes' in data):
                converted_data = ConvertedData(data['deviceName'], data.get('deviceType', 'default'))
                if "timeseries" in data:
                    converted_data.add_to_telemetry(data['timeseries'])
                if "telemetry" in data:
                    converted_data.add_to_telemetry(data['telemetry'])

                converted_data.add_to_attributes(data['attributes'])
                return converted_data
            else:
                self._log.error("Failed to convert SQL data to TB format: no configuration provided")
                return ConvertedData(None, None)

        device_report_strategy = None
        try:
            device_report_strategy = ReportStrategyConfig(config if isinstance(config, dict) else None)
        except ValueError as e:
            self._log.trace("Device report strategy configuration is not provided: %s", str(e))

        converted_data = ConvertedData(None, None)
        for datatype in self.DATATYPES:
            for config_item in config.get(datatype, []):
                try:
                    full_key = None
                    value = None
                    if isinstance(config_item, str):
                        if isinstance(data, dict):
                            for key, value in data.items():
                                full_key = TBUtility.convert_key_to_datapoint_key(key, device_report_strategy, {}, self._log)
                                if full_key:
                                    if datatype == TIMESERIES_PARAMETER:
                                        converted_data.add_to_telemetry(TelemetryEntry({full_key: value}, data.get('ts')))
                                    elif datatype == ATTRIBUTES_PARAMETER:
                                        converted_data.add_to_attributes(full_key, value)
                            continue
                        elif isinstance(data, list):
                            for item in data:
                                for key, value in item.items():
                                    full_key = TBUtility.convert_key_to_datapoint_key(key, device_report_strategy, {}, self._log)
                                    if full_key:
                                        if datatype == TIMESERIES_PARAMETER:
                                            converted_data.add_to_telemetry(TelemetryEntry({full_key: value}, item.get('ts')))
                                        elif datatype == ATTRIBUTES_PARAMETER:
                                            converted_data.add_to_attributes(full_key, value)
                            continue
                    elif isinstance(config_item, dict):
                        if "nameExpression" in config_item:
                            name = eval(config_item["nameExpression"], globals(), data)
                        else:
                            name = config_item["name"]
                        full_key = TBUtility.convert_key_to_datapoint_key(name, device_report_strategy, config_item, self._log)

                        if "column" in config_item:
                            value = data[config_item["column"]]
                        elif "value" in config_item:
                            value = eval(config_item["value"], globals(), data)
                        else:
                            StatisticsService.count_connector_message(self._log.name, 'convertersMsgDropped')
                            self._log.error("Failed to convert SQL data to TB format: no column/value configuration item")
                            continue

                    if not full_key:
                        StatisticsService.count_connector_message(self._log.name, 'convertersMsgDropped')
                        self._log.error("Failed to convert SQL data to TB format: no full key")
                        continue

                    if datatype == TIMESERIES_PARAMETER:
                        converted_data.add_to_telemetry(TelemetryEntry({full_key: value}, data.get('ts')))
                    elif datatype == ATTRIBUTES_PARAMETER:
                        converted_data.add_to_attributes(full_key, value)

                    else:
                        StatisticsService.count_connector_message(self._log.name, 'convertersMsgDropped')
                        self._log.error("Failed to convert SQL data to TB format: unexpected configuration type '%s'",
                                  type(config_item))
                except Exception as e:
                    StatisticsService.count_connector_message(self._log.name, 'convertersMsgDropped')
                    self._log.error("Failed to convert SQL data to TB format: %s", str(e))

        return converted_data



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

import io
import re
from ftplib import FTP, FTP_TLS
from queue import Queue
from random import choice
from re import fullmatch
from string import ascii_lowercase
from threading import Thread
from time import perf_counter as timer, sleep

import simplejson

from thingsboard_gateway.connectors.ftp.backward_compatibility_adapter import FTPBackwardCompatibilityAdapter
from thingsboard_gateway.connectors.ftp.file import File
from thingsboard_gateway.connectors.ftp.ftp_uplink_converter import FTPUplinkConverter
from thingsboard_gateway.connectors.ftp.path import Path
from thingsboard_gateway.gateway.entities.converted_data import ConvertedData
from thingsboard_gateway.gateway.statistics.decorators import CollectAllReceivedBytesStatistics
from thingsboard_gateway.gateway.statistics.statistics_service import StatisticsService
from thingsboard_gateway.tb_utility.tb_logger import init_logger

from thingsboard_gateway.connectors.connector import Connector
from thingsboard_gateway.tb_utility.tb_utility import TBUtility


class FTPConnector(Connector, Thread):
    def __init__(self, gateway, config, connector_type):
        super().__init__()
        self.statistics = {'MessagesReceived': 0,
                           'MessagesSent': 0}
        using_old_config_format_detected = FTPBackwardCompatibilityAdapter.is_old_config_format(config)
        if using_old_config_format_detected:
            self.config = FTPBackwardCompatibilityAdapter(config).convert()
            self.__id = self.config.get("id")
        else:
            self.config = config
            self.__id = self.config.get("id")

        self._connector_type = connector_type
        self.__gateway = gateway
        self.security = {**self.config['parameters']['security']} if self.config['parameters']['security'][
                                                                         'type'] == 'basic' else {
            'username': 'anonymous', "password": 'anonymous@'}
        self.__tls_support = self.config['parameters'].get("TLSSupport", False)
        self.name = self.config.get("name", "".join(choice(ascii_lowercase) for _ in range(5)))
        self.__log = init_logger(self.__gateway, self.name, self.config.get('logLevel', 'INFO'),
                                 enable_remote_logging=self.config.get('enableRemoteLogging', False),
                                 is_connector_logger=True)
        self.__converter_log = init_logger(self.__gateway, self.name + '_converter',
                                           self.config.get('logLevel', 'INFO'),
                                           enable_remote_logging=self.config.get('enableRemoteLogging', False),
                                           is_converter_logger=True, attr_name=self.name)
        self.daemon = True
        self.__stopped = False
        self.__requests_in_progress = []
        self.__convert_queue = Queue(1000000)
        self.__attribute_updates = []
        self.__fill_attributes_update()
        self._connected = False
        self.__rpc_requests = []
        self.start_time = timer()
        self.__fill_rpc_requests()
        self.host = self.config['parameters']['host']
        self.port = self.config['parameters'].get('port', 21)
        self.__ftp = FTP_TLS if self.__tls_support else FTP
        self.paths = [
            Path(
                path=obj['path'],
                read_mode=obj['readMode'],
                telemetry=obj['timeseries'],
                device_name=obj['devicePatternName'],
                attributes=obj['attributes'],
                txt_file_data_view=obj.get('txtFileDataView', 'TABLE'),
                with_sorting_files=obj.get('with_sorting_files', True),
                poll_period=obj.get('pollPeriod', 60),
                max_size=obj.get('maxFileSize', 5),
                delimiter=obj.get('delimiter', ','),
                device_type=obj.get('devicePatternType', 'Device'),
                report_strategy=obj.get('reportStrategy')
            )
            for obj in self.config['paths']
        ]
        self.__log.info("FTP Connector started with %s and %d", self.host, self.port)

    def open(self):
        self.__stopped = False
        self.start()

    def run(self):
        self.__log.debug('Starting connector loop')
        try:
            while not self.__stopped:
                with self.__ftp() as ftp:
                    self.__connect(ftp)

                    if self._connected:
                        for path in self.paths:
                            path.find_files(ftp)

                        while True:
                            sleep(.2)
                            self.__process_paths(ftp)
                            if self.__stopped:
                                break
        except Exception as e:
            self.__log.error("Unexpected exception in loop for %s %d with error %r", self.host, self.port, str(e))
            self.__log.debug("Error:", exc_info=e)
            try:
                self.close()
            except Exception as e:
                self.__log.error(
                    "Can not close the connection for %s %d with error %r", self.host, self.port, str(e))
                self.__log.debug("Error:", exc_info=e)
        while True:
            if self.__stopped:
                break

    def __connect(self, ftp):
        self.__log.debug("Connecting to ftp server on %s:%d", self.host, self.port)
        try:
            ftp.connect(self.host, self.port)

            if isinstance(ftp, FTP_TLS):
                ftp.sendcmd('USER ' + self.security['username'])
                ftp.sendcmd('PASS ' + self.security['password'])
                ftp.prot_p()
                self.__log.info('Data protection level set to "private"')
            else:
                ftp.login(self.security['username'], self.security['password'])
                self.__log.info("Logged in as %s", str(self.security['username']))


        except Exception as e:
            self.__log.error("Connection failed to %s:%d: due to %r", self.host, self.port, str(e))
            self.__log.debug("Error:", exc_info=e)
            sleep(10)
        else:
            self._connected = True
            self.__log.info("Connected to FTP server to %s:%d", self.host, self.port)

    def __process_paths(self, ftp):
        for path in self.paths:
            time_point = timer()
            if time_point - path.last_polled_time >= path.poll_period or path.last_polled_time == 0:
                configuration = path.config
                converter = FTPUplinkConverter(configuration, self.__converter_log)
                path.last_polled_time = time_point

                if '*' in path.path:
                    path.find_files(ftp)
                    self.__log.trace("Found %d for pattern %s", len(path.files), path.path)

                for file in path.files:
                    current_hash = file.get_current_hash(ftp)
                    if ((file.has_hash() and current_hash != file.hash)
                        or not file.has_hash()) and file.check_size_limit(ftp):
                        file.set_new_hash(current_hash)

                        handle_stream = io.BytesIO()
                        self.__log.trace("Retrieving file %s...", file.path_to_file)

                        ftp.retrbinary('RETR ' + file.path_to_file, handle_stream.write)

                        handled_str = str(handle_stream.getvalue(), 'UTF-8')
                        handled_array = handled_str.split('\n')

                        StatisticsService.count_connector_message(self.name,
                                                                  stat_parameter_name='connectorMsgsReceived')
                        StatisticsService.count_connector_bytes(self.name, handled_str,
                                                                stat_parameter_name='connectorBytesReceived')

                        convert_conf = {'file_ext': file.path_to_file.split('.')[-1]}

                        self.__log.trace("Processing data from %s file", file.path_to_file)

                        if convert_conf['file_ext'] == 'json':
                            json_data = simplejson.loads(handled_str)
                            if isinstance(json_data, list):
                                for obj in json_data:
                                    converted_data = converter.convert(convert_conf, obj)

                                    if converted_data:
                                        self.__log.info(
                                            'Converted data for device %s with type %s, attributes: %s, telemetry: %s',
                                            converted_data.device_name, converted_data.device_type,
                                            converted_data.attributes_datapoints_count,
                                            converted_data.telemetry_datapoints_count)

                                        self.__log.debug('Converted data: %s', converted_data)
                                        self.__send_data(converted_data)
                            else:
                                converted_data = converter.convert(convert_conf, json_data)

                                if converted_data:
                                    self.__log.info(
                                        'Converted data for device %s with type %s, attributes: %s, telemetry: %s',
                                        converted_data.device_name, converted_data.device_type,
                                        converted_data.attributes_datapoints_count,
                                        converted_data.telemetry_datapoints_count)

                                    self.__log.debug('Converted data: %s', converted_data)
                                    self.__send_data(converted_data)
                        else:
                            cursor = file.cursor or 0

                            for (index, line) in enumerate(handled_array):
                                if index == 0 and not path.txt_file_data_view == 'SLICED':
                                    convert_conf['headers'] = line.split(path.delimiter)
                                else:
                                    if file.read_mode == File.ReadMode.PARTIAL and index >= cursor:
                                        converted_data = converter.convert(convert_conf, line)
                                        if index + 1 == len(handled_array):
                                            file.cursor = index
                                    else:
                                        converted_data = converter.convert(convert_conf, line)

                                    if converted_data:
                                        self.__log.info(
                                            'Converted data for device %s with type %s, attributes: %s, telemetry: %s',
                                            converted_data.device_name, converted_data.device_type,
                                            converted_data.attributes_datapoints_count,
                                            converted_data.telemetry_datapoints_count)

                                        self.__log.debug('Converted data: %s', converted_data)
                                        self.__send_data(converted_data)

                        handle_stream.close()

    def __send_data(self, converted_data: ConvertedData):
        if (converted_data and
                (converted_data.telemetry_datapoints_count > 0 or
                 converted_data.attributes_datapoints_count > 0)):
            self.__gateway.send_to_storage(self.name, self.get_id(), converted_data)
            self.statistics['MessagesSent'] = self.statistics['MessagesSent'] + 1
            self.__log.debug("Data being sent to ThingsBoard: %s", converted_data)

    def close(self):
        self.__stopped = True
        for path in self.paths:
            for file in path.files:
                file._hash = None
        self.__log.info('FTP Connector stopped.')
        self.__log.stop()

    def get_name(self):
        return self.name

    def get_id(self):
        return self.__id

    def get_type(self):
        return self._connector_type

    def is_connected(self):
        return self._connected

    def is_stopped(self):
        return self.__stopped

    def __fill_attributes_update(self):
        for attribute_request in self.config.get('requestsMapping', {}).get('attributeUpdates', []):
            self.__attribute_updates.append(attribute_request)

    @staticmethod
    def __is_json(data):
        try:
            simplejson.loads(data)
        except ValueError:
            return False
        return True

    @CollectAllReceivedBytesStatistics(start_stat_type='allReceivedBytesFromTB')
    def on_attributes_update(self, content):
        self.__log.debug("Processing an attribute update from %r", content)
        try:
            for attribute_request in self.__attribute_updates:
                if fullmatch(attribute_request["deviceNameFilter"], content["device"]):
                    attribute_key, attribute_value = content['data'].popitem()
                    self.__log.info("Found an attribute key for %s and value %s", attribute_key, attribute_value)

                    path_str = attribute_request['path'].replace('${attributeKey}', attribute_key).replace(
                        '${attributeValue}', attribute_value)
                    path = Path(path=path_str, device_name=content['device'], attributes=[], telemetry=[],
                                delimiter=',', txt_file_data_view='')

                    data_expression = (attribute_request['valueExpression']
                                       .replace('${attributeKey}', attribute_key)
                                       .replace('${attributeValue}', attribute_value))

                    with self.__ftp() as ftp:
                        self.__connect(ftp)
                        path.find_files(ftp)
                        for file in path.files:
                            if file.path_to_file.split('.')[-1] == 'json':
                                json_data = data_expression.replace("'", '"')
                                if self.__is_json(json_data):
                                    io_stream = io.BytesIO(str.encode(json_data))
                                    ftp.storbinary('STOR ' + file.path_to_file, io_stream)
                                    io_stream.close()
                                else:
                                    self.__log.error("Invalid json data in attribute update %s", json_data)
                                    self.__log.debug("Error:", exc_info=True)
                            else:
                                if attribute_request['writingMode'] == 'OVERRIDE':
                                    io_stream = self._get_io_stream(data_expression)
                                    ftp.storbinary('STOR ' + file.path_to_file, io_stream)
                                    io_stream.close()
                                    self.__log.info("Successfully process attribute update %s on override request", attribute_key)
                                else:
                                    handle_stream = io.BytesIO()
                                    ftp.retrbinary('RETR ' + file.path_to_file, handle_stream.write)
                                    converted_data = str(handle_stream.getvalue(), 'UTF-8')
                                    handle_stream.close()

                                    io_stream = io.BytesIO(str.encode(str(converted_data + '\n' + data_expression)))
                                    ftp.storbinary('STOR ' + file.path_to_file, io_stream)
                                    io_stream.close()
                                    self.__log.info("Successfully process attribute update for %s", attribute_key)

        except Exception as e:
            self.__log.error("Failed to process attribute update with error %r", str(e))
            self.__log.debug("Error:", exc_info=e)

    @CollectAllReceivedBytesStatistics('allBytesSentToDevices')
    def _get_io_stream(self, data_expression):
        return io.BytesIO(str.encode(data_expression))

    def __fill_rpc_requests(self):
        for rpc_request in self.config.get('requestsMapping', {}).get("serverSideRpc", []):
            self.__rpc_requests.append(rpc_request)

    @CollectAllReceivedBytesStatistics(start_stat_type='allReceivedBytesFromTB')
    def server_side_rpc_handler(self, content):
        try:
            self.__log.debug("Handling incoming server-side RPC with %s", content)

            if content.get('data') is None:
                content['data'] = {'params': content['params'], 'method': content['method'], 'id': content['id']}

            rpc_method = content['data']['method']

            # check if RPC type is connector RPC
            try:
                (connector_type, rpc_method_name) = rpc_method.split('_')
                if connector_type == self._connector_type:
                    value_expression = content['data']['params']['valueExpression']
                    converted_data, success_sent = self.__process_rpc(rpc_method_name, value_expression)
                    self.__send_rpc_reply({}, content, converted_data, success_sent)
                    return
            except ValueError:
                pass

            # check if RPC method is reserved get/set
            if rpc_method == 'get' or rpc_method == 'set':
                params = {}
                for param in content['data']['params'].split(';'):
                    try:
                        (key, value) = param.split('=')
                    except ValueError:
                        continue

                    if key and value:
                        params[key] = value

                rpc_method = 'write' if rpc_method == 'set' else 'read'

                if rpc_method == 'read':
                    value_expression = params.get('filePath')
                else:
                    value_expression = params.get('filePath') + ';' + params.get('value')

                converted_data, success_sent = self.__process_rpc(rpc_method, value_expression)
                self.__send_rpc_reply({}, content, converted_data, success_sent)
                self.__log.info("Successfully sent RPC request to FTP for %s rpc method", rpc_method)
                return

            for rpc_request in self.__rpc_requests:
                if not fullmatch(rpc_request['deviceNameFilter'], content['device']):
                    continue
                if fullmatch(rpc_request['methodFilter'], rpc_method):
                    params_field_expression = rpc_request['valueExpression']

                    value_expression_key = TBUtility.get_value(params_field_expression, content['data'], get_tag=True)
                    value_expression = content['data'][value_expression_key]
                    converted_data, success_sent = self.__process_rpc(rpc_method, value_expression)

                    self.__send_rpc_reply(rpc_request, content, converted_data, success_sent)
                    self.__log.info("Successfully sent RPC request to FTP for %s rpc method", rpc_method)

        except Exception as e:
            self.__log.error(
                "Failed to perform incoming server side RPC for content %s and rpc method due to %r", str(e))
            self.__log.debug("Error:", exc_info=e)

    def __process_rpc(self, method, value_expression):
        self.__log.info("Called __process_rpc, %r %r", method, value_expression)
        with self.__ftp() as ftp:
            if not self._connected or not ftp.sock:
                self.__connect(ftp)

            converted_data = None
            success_sent = None
            if method == 'write':
                try:
                    arr = re.sub("'", '', value_expression).split(';')
                    io_stream = self._get_io_stream(arr[1])
                    ftp.storbinary('STOR ' + arr[0], io_stream)
                    io_stream.close()
                    success_sent = True
                    converted_data = {"result": {"value": arr[1]}} if len(arr[1]) < 80 else {"result": True}
                    self.__log.info("The value %s is written to %s", arr[1], arr[0])
                except Exception as e:
                    self.__log.error("Can not process for method write due to %r", str(e))
                    self.__log.debug("Error:", exc_info=e)
                    converted_data = '{"error": "' + str(e) + '"}'
            else:
                handle_stream = io.BytesIO()
                ftp.retrbinary('RETR ' + value_expression, handle_stream.write)
                converted_data = str(handle_stream.getvalue(), 'UTF-8')
                handle_stream.close()

            return converted_data, success_sent

    def __send_rpc_reply(self, rpc_request, content, converted_data, success_sent):
        if content.get('device') and fullmatch(rpc_request.get('deviceNameFilter', ''), content.get('device')):
            self.__gateway.send_rpc_reply(device=content["device"], req_id=content["data"]["id"],
                                          success_sent=success_sent, content={'result': converted_data})
        elif content.get('device'):
            self.__gateway.send_rpc_reply(device=content["device"], req_id=content["data"]["id"],
                                          success_sent=success_sent, content={'result': converted_data})
        else:
            return converted_data

    def get_config(self):
        return self.config



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from asyncio import Queue, QueueEmpty, sleep
from typing import List
from bacpypes3.ipv4.app import NormalApplication, ForeignApplication
from bacpypes3.local.device import DeviceObject
from bacpypes3.pdu import Address, IPv4Address
from bacpypes3.primitivedata import ObjectIdentifier, Unsigned
from bacpypes3.basetypes import PropertyIdentifier, ReadAccessSpecification
from bacpypes3.vendor import get_vendor_info
from bacpypes3.constructeddata import SequenceOf, Array
from bacpypes3.apdu import (
    APDU,
    AbortPDU,
    ComplexAckPDU,
    ErrorPDU,
    RejectPDU,
    SimpleAckPDU,
    ErrorRejectAbortNack,
    ReadPropertyMultipleRequest,
    ReadPropertyMultipleACK
)
from bacpypes3.comm import bind

from thingsboard_gateway.connectors.bacnet.application_service_access_point import ApplicationServiceAccessPoint
from thingsboard_gateway.connectors.bacnet.entities.device_object_config import DeviceObjectConfig


class Application(NormalApplication, ForeignApplication):
    def __init__(self, device_object_config: DeviceObjectConfig, indication_callback, logger, is_foreign_application: bool = False):
        self.__stopped = False

        self.__device_object_config = device_object_config
        self.__device_object = DeviceObject(**self.__device_object_config.device_object_config)

        if not is_foreign_application:
            NormalApplication.__init__(
                self, self.__device_object, Address(self.__device_object_config.address)
            )
        else:
            ForeignApplication.__init__(
                self, self.__device_object, Address(self.__device_object_config.address)
            )

        self.asap = ApplicationServiceAccessPoint(self.device_object, self.device_info_cache)
        bind(self, self.asap, self.nsap)

        self.__log = logger
        self.__indication_callback = indication_callback
        self.__confirmation_queue = Queue(1_000_000)
        self.__is_foreign_application = is_foreign_application

    def register_foreign_device(self, address: IPv4Address, ttl: int) -> None:
        if self.__is_foreign_application:
            self.__log.debug(f"(register_foreign_device) Registering foreign device")
            self.register(address, ttl)
        else:
            self.__log.error(f"(register_foreign_device) Registering foreign device not possible because application is no foreign application")

    def unregister_foreign_device(self) -> None:
        if self.__is_foreign_application:
            self.__log.debug(f"(unregister_foreign_device) Unregistering foreign device")
            self.unregister()
        else:
            self.__log.debug(f"(unregister_foreign_device) Unregistering foreign device not needed because application is no foreign application")

    def close(self):
        self.__stopped = True
        self.unregister_foreign_device()
        super().close()

    async def indication(self, apdu) -> None:
        self.__log.debug(f"(indication) Received APDU: {apdu}")
        await super().indication(apdu)
        self.__indication_callback(apdu)

    async def confirmation(self, apdu: APDU) -> None:
        self.__confirmation_queue.put_nowait(apdu)

    async def confirmation_handler(self):
        while not self.__stopped:
            try:
                apdu = self.__confirmation_queue.get_nowait()

                if apdu.pduSource is None:
                    self.__log.warning("Received APDU without source address: %s", apdu)
                    return

                pdu_source = apdu.pduSource
                if pdu_source not in self._requests:
                    return

                requests = self._requests[pdu_source]
                for indx, (request, future) in enumerate(requests):
                    if request.apduInvokeID == apdu.apduInvokeID:
                        break
                else:
                    return

                if isinstance(apdu, (SimpleAckPDU, ComplexAckPDU)):
                    future.set_result(apdu)
                elif isinstance(apdu, (ErrorPDU, RejectPDU, AbortPDU)):
                    future.set_exception(apdu)
                else:
                    raise TypeError("apdu")
            except QueueEmpty:
                await sleep(0.1)
            except Exception as e:
                self.__log.error("APDU confirmation error: %s", e)

    async def do_who_is(self, device_address):
        try:
            devices = await self.who_is(address=Address(device_address),
                                        timeout=self.__device_object_config.device_discovery_timeout)
            if len(devices):
                return devices[0]

        except ValueError as e:
            self.__log.error("Incorrect data format is used %s", str(e))
            self.__log.debug("Error %s", e, exc_info=e)

        except Exception as e:
            self.__log.error("An unexpected error occurred while device look up process %s", str(e))
            self.__log.debug("Error %s", e, exc_info=e)

    async def get_object_identifiers_with_segmentation(self, device) -> List[ObjectIdentifier]:
        object_list = await self.__send_request_wrapper(self.read_property,
                                                        err_msg=f"Failed to read {device.details.identifier} object-list",  # noqa
                                                        address=device.details.address,
                                                        objid=device.details.identifier,
                                                        prop='objectList')

        if object_list is None:
            return []

        return list(filter(lambda obj_id: str(obj_id[0]) != 'device', object_list))

    async def get_object_identifiers_without_segmentation(self, device, index_to_read=None) -> List[ObjectIdentifier]:
        object_list = []
        success_indexes = []

        if index_to_read is None:
            object_list_length = await self.__send_request_wrapper(self.read_property,
                                                                   err_msg=f"Failed to read {device.details.identifier} object-list length",  # noqa
                                                                   address=device.details.address,
                                                                   objid=device.details.identifier,
                                                                   prop='objectList',
                                                                   array_index=0)
            if object_list_length is None:
                return []

            device.details.objects_len = object_list_length
            index_to_read = range(object_list_length)

        for i in index_to_read:
            object_identifier = await self.__send_request_wrapper(self.read_property,
                                                                  err_msg=f"Failed to read {device.details.identifier} object-list[{i + 1}]",  # noqa
                                                                  address=device.details.address,
                                                                  objid=device.details.identifier,
                                                                  prop='objectList',
                                                                  array_index=i + 1)
            if object_identifier is None:
                device.details.failed_to_read_indexes = i
                continue

            if object_identifier[0].__str__() == 'device':
                continue

            object_list.append(object_identifier)
            success_indexes.append(i)

        for i in success_indexes:
            device.details.sucess_read_for(i)

        return object_list

    async def read_multiple_objects(self, device, object_list):
        if len(object_list) == 0:
            self.__log.warning("%s no objects to read", device.details.object_id)
            return []

        read_access_specifications = self.__get_read_access_specifications(object_list, device.details.vendor_id)
        if len(read_access_specifications) == 0:
            self.__log.warning("no read access specifications")
            return []

        request = ReadPropertyMultipleRequest(
            listOfReadAccessSpecs=SequenceOf(ReadAccessSpecification)(
                read_access_specifications
            ),
            destination=Address(device.details.address),
        )

        result = await self.__send_request_wrapper(self.request,
                                                   err_msg=f"Failed to read {device.details.object_id} objects",
                                                   apdu=request)

        if not isinstance(result, ReadPropertyMultipleACK):
            self.__log.error("Invalid response type: %s", type(result))
            return []

        decoded_result = self.decode_tag_list(result, device.details.vendor_id)

        return decoded_result

    async def get_device_objects(self, device, with_all_properties=False, index_to_read=None):
        if device.details.is_segmentation_supported():
            object_list = await self.get_object_identifiers_with_segmentation(device)
        else:
            object_list = await self.get_object_identifiers_without_segmentation(device,
                                                                                 index_to_read=index_to_read)

        if len(object_list) == 0:
            self.__log.warning("%s no objects to read", device.details.object_id)
            return

        object_list = [{'objectId': obj, 'propertyId': 'object-name' if not with_all_properties else 'all'}
                       for obj in object_list]

        return ObjectIterator(self, device, object_list, self.read_multiple_objects)

    async def get_device_values(self, device):
        return ObjectIterator(self,
                              device,
                              device.uplink_converter_config.objects_to_read,
                              self.read_multiple_objects)

    def __get_read_access_specifications(self, object_list, vendor_id):
        read_access_specifications = []
        vendor_info = get_vendor_info(vendor_id)

        for object in object_list:
            try:
                object_id = object['objectId']
                if not isinstance(object_id, ObjectIdentifier):
                    obj_str = f"{object['objectType']},{object_id}"
                    object_id = ObjectIdentifier(obj_str)

                object_class = vendor_info.get_object_class(object_id[0])
                if object_class is None:
                    self.__log.warning(f"unknown object type: {object_id}, {object_class}")
                    continue

                properties = []
                if not isinstance(object['propertyId'], set):
                    object['propertyId'] = {object['propertyId']}

                for prop in object['propertyId']:
                    property_identifier = PropertyIdentifier(prop)

                    properties.append(property_identifier)

                ras = ReadAccessSpecification(
                    objectIdentifier=object_id,
                    listOfPropertyReferences=properties,
                )

                read_access_specifications.append(ras)
            except Exception as e:
                self.__log.error("failed to create read access specification for %s: %s", object_id, e)
                continue

        return read_access_specifications

    async def get_device_name(self, address: Address, object_id: ObjectIdentifier):
        device_name = await self.__send_request_wrapper(self.read_property,
                                                        err_msg=f"Failed to read {address} device name",
                                                        address=address,
                                                        objid=object_id,
                                                        prop='objectName')

        return device_name

    async def get_router_info(self, device_address: Address):
        router_info = {}

        # who-is method direct call doesn't trigger the indication callback
        router = await self.__send_request_wrapper(self.who_is,
                                                   err_msg=f"Failed to discover router info for {device_address}",
                                                   address=device_address)
        if router is not None:
            router = router[0] if isinstance(router, list) else router
        else:
            return None

        router_name = await self.get_device_name(router.pduSource, router.iAmDeviceIdentifier)
        if router_name is None:
            return None

        try:
            router_info['routerVendorId'] = router.vendorID
            router_info['routerId'] = router.iAmDeviceIdentifier[1]
            router_info['routerAddress'] = str(router.pduSource)
            router_info['routerName'] = router_name
        except Exception as e:
            self.__log.error("Failed to parse router %s info: %s", device_address, e)
            return None

        return router_info

    async def __send_request_wrapper(self, func, err_msg=None, *args, **kwargs):
        """
        Helper method to send a request and handle exceptions.
        """

        try:
            return await func(*args, **kwargs)
        except AbortPDU as e:
            self.__log.warning("(Request aborted) %s: %s", err_msg, e)
        except ErrorRejectAbortNack as e:
            self.__log.warning("(Request rejected) %s: %s", err_msg, e)
        except ErrorPDU as e:
            self.__log.error("(Error in request) %s: %s", err_msg, e)
        except Exception as e:
            self.__log.error("(Unexpected error in request) %s: %s", err_msg, e)

        return None

    def decode_tag_list(self, tag_list, vendor_id):
        vendor_info = get_vendor_info(vendor_id)
        result_list = []

        for read_access_result in tag_list.listOfReadAccessResults:
            object_identifier = read_access_result.objectIdentifier
            object_class = vendor_info.get_object_class(object_identifier[0])

            for read_access_result_element in read_access_result.listOfResults:
                try:
                    property_identifier = read_access_result_element.propertyIdentifier
                    property_array_index = read_access_result_element.propertyArrayIndex
                    read_result = read_access_result_element.readResult

                    if read_result.propertyAccessError:
                        result_list.append(
                            (
                                object_identifier,
                                property_identifier,
                                property_array_index,
                                read_result.propertyAccessError,
                            )
                        )
                        continue

                    property_type = object_class.get_property_type(property_identifier)
                    if property_type is None:
                        self.__log.warning("%r not supported", property_identifier)

                        result_list.append(
                            (
                                object_identifier,
                                property_identifier,
                                property_array_index,
                                None,
                            )
                        )
                        continue

                    if issubclass(property_type, Array):
                        if property_array_index is None:
                            pass
                        elif property_array_index == 0:
                            property_type = Unsigned
                        else:
                            property_type = property_type._subtype

                    property_value = read_result.propertyValue.cast_out(property_type)

                    result_list.append(
                        (
                            object_identifier,
                            property_identifier,
                            property_array_index,
                            property_value,
                        )
                    )
                except Exception as e:
                    self.__log.error('failed to decode read access result: %s', e)
                    continue

        return result_list


class ObjectIterator:
    def __init__(self, app, device, object_list, func):
        self.app = app
        self.items = object_list
        self.device = device
        self.limit = self.get_limit()
        self.func = func
        self.index = 0

    async def get_next(self):
        if self.index >= len(self.items):
            return [], {}, True

        end_index = self.index + self.limit
        result = self.items[self.index:end_index]
        self.index = end_index
        finished = self.index >= len(self.items)

        r = await self.func(self.device, result)

        return r, result, finished

    def get_limit(self):
        max_prop_count = 0

        for item in self.items:
            cur_prop_count = len(item['propertyId']) if isinstance(item['propertyId'], set) else 1
            if item['propertyId'] == 'all':
                max_prop_count = 6
                break

            if cur_prop_count > max_prop_count:
                max_prop_count = cur_prop_count

        return int(self.device.details.get_max_apdu_count() / max_prop_count) if max_prop_count > 0 else 1



#     Copyright 2025. ThingsBoard
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.

from os import path


class StorageSettings:
    def __init__(self, config, enable_validation=True):
        self.enable_validation = enable_validation
        self.data_file_path = config.get("data_file_path", "./")
        self.messages_ttl_check_in_hours = (
                config.get("messages_ttl_check_in_hours", 1) * 3600
        )
        self.messages_ttl_in_days = config.get("messages_ttl_in_days", 7)
        self.max_read_records_count = config.get("max_read_records_count", 1000)
        self.batch_size = config.get("writing_batch_size", 1000)
        self.directory_path = path.dirname(self.data_file_path)
        self.db_file_name = "data.db"
        self.size_limit = config.get("size_limit", 1024)
        self.max_db_amount = config.get("max_db_amount", 10)
        self.oversize_check_period = config.get("oversize_check_period", 1)
        self.warnings = self.validate_settings()

    def validate_settings(self):
        warnings = []
        if path.basename(self.data_file_path):
            warnings.append(
                "You can not specify the file name, using default file name: data.db or make sure you set slash at the end of the path;")

        if self.size_limit < 1 and self.enable_validation:
            self.size_limit = 1
            warnings.append("The size limit is too small - using the minimum value 1 MB;")

        if self.oversize_check_period < 1 and self.enable_validation:
            self.oversize_check_period = 1
            warnings.append("The oversize check period is too small - using the minimum value 1 minute;")

        return warnings

